/*
 * Copyright (C) 2002  Benjie Chen (benjie@lcs.mit.edu),
 *                     Massachusetts Institute of Technology
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 *  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 *  DEALINGS IN THE SOFTWARE.
 */

#include "sfscd_prot.h"
#include "cfsa_prot.h"
#include "nfs3close_prot.h"
#include "nfs3_nonnul.h"
#include "rxx.h"

#include "nfstrans.h"
#include "nfsxfattr.h"

#include "dhash.h"
#include "ranges.h"
#include "log.h"
#include "logiter.h"

#include "cfsrwcd.h"
#include "fcache.h"

extern int ncacheflush;
extern int nsyncwrite;

//
// see README for notes
//

aiod* cfsrwcd::a = New aiod (2);

str
cfsrwcd::fh2fn(nfs_fh3 fh)
{
  strbuf n;
  unsigned char x = 0;
  char *c = fh.data.base();
  for (unsigned i=0; i<fh.data.size(); i++)
    x = x ^ ((unsigned char)(*(c+i)));
  n << _cdir << "/" << (((unsigned int)x)%254) << "/"
    << armor32(fh.data.base(), fh.data.size());
  return n;
}

void
cfsrwcd::run_rpcs (file_cache *e)
{
  vec<nfscall *> rpcs;
  for (unsigned i=0; i<e->rpcs.size(); i++)
    rpcs.push_back (e->rpcs[i]);
  e->rpcs.clear();
  for (unsigned i=0; i<rpcs.size(); i++)
    dispatch_nfs (rpcs[i]);
}

void
cfsrwcd::fill_done (file_cache *e, bool done, bool ok)
{
  assert(e->is_fetch());
  if (done) {
    if (ok)
      e->idle();
    else
      e->error();
  }
  run_rpcs (e);
}

void
cfsrwcd::read_from_cache (nfscall *sbp, file_cache *e)
{
  if (e->afh == 0) {
    str fn = fh2fn(e->fh);
    cfsrwcd::a->open (fn, O_RDWR, 0666,
	              wrap (this, &cfsrwcd::read_from_cache_open, sbp, e));
  }
  else
    read_from_cache_open (sbp, e, e->afh, 0);
}

void
cfsrwcd::read_from_cache_open (nfscall *sbp, file_cache *e,
                               ptr<aiofh> afh, int err)
{
  if (!afh) {
    warn << "read_from_cache: open failed: " << err << "\n";
    sbp->reject (SYSTEM_ERR);
    return;
  }
  e->afh = afh;

  read3args *a = sbp->template getarg<read3args> ();
  WARN(2) << "read from cache at " << a->offset << "+" << a->count << "\n";

  ptr<aiobuf> buf = cfsrwcd::a->bufalloc (a->count+1);
  if (!buf) {
    cfsrwcd::a->bufwait
      (wrap (this, &cfsrwcd::read_from_cache_open, sbp, e, e->afh, 0));
    return;
  }

  e->afh->read (a->offset, buf,
                wrap (this, &cfsrwcd::read_from_cache_read, sbp, e));
}

void
cfsrwcd::read_from_cache_read (nfscall *sbp, file_cache *e,
                               ptr<aiobuf> buf, ssize_t sz, int err)
{
  if (err) {
    warn << "read_from_cache: read failed: " << err << "\n";
    sbp->reject (SYSTEM_ERR);
  }

  read3args *a = sbp->template getarg<read3args> ();
  int x = ((unsigned)sz) > a->count ? a->count : sz;
  read3res res(NFS3_OK);
  res.resok->count = x;
  res.resok->data.setsize(x);
  memcpy(res.resok->data.base(), buf->base(), x);
  res.resok->eof = (((unsigned)sz) <= a->count);
  res.resok->file_attributes.set_present (true);
  *res.resok->file_attributes.attributes = e->fa;
  sbp->reply (&res);
}

void
cfsrwcd::write_to_cache (nfscall *sbp, file_cache *e)
{
  e->outstanding_op ();

  if (e->afh == 0) {
    str fn = fh2fn(e->fh);
    cfsrwcd::a->open (fn, O_RDWR, 0666,
	              wrap (this, &cfsrwcd::write_to_cache_open, sbp, e));
  }
  else
    write_to_cache_open (sbp, e, e->afh, 0);
}

void
cfsrwcd::write_to_cache_open (nfscall *sbp, file_cache *e,
                              ptr<aiofh> afh, int err)
{
  if (!afh) {
    warn << "write_to_cache: open failed: " << err << "\n";
    e->error ();
    sbp->reject (SYSTEM_ERR);
    run_rpcs (e);
    return;
  }

  write3args *a = sbp->template getarg<write3args> ();

  e->afh = afh;
  assert (e->is_idle() || e->is_dirty());

  ptr<aiobuf> buf = cfsrwcd::a->bufalloc (a->count);
  if (!buf) {
    cfsrwcd::a->bufwait
      (wrap (this, &cfsrwcd::write_to_cache_open, sbp, e, e->afh, 0));
    return;
  }
 
  memmove(buf->base (), a->data.base (), a->count);
  e->afh->write (a->offset, buf,
                 wrap (this, &cfsrwcd::write_to_cache_write, sbp, e));
}

void
cfsrwcd::write_to_cache_write (nfscall *sbp, file_cache *e,
                               ptr<aiobuf> buf, ssize_t sz, int err)
{
  write3args *a = sbp->template getarg<write3args> ();
  e->outstanding_op_done ();

  if (err || (unsigned)sz != a->count) {
    warn << "write_to_cache: write failed: " << err << "\n";
    e->error ();
    sbp->reject (SYSTEM_ERR);
  }
  else {
    e->dirty ();

    if (a->offset+sz > e->fa.size)
      e->fa.size = a->offset+sz;
  
    if (a->stable == UNSTABLE) {
      // mark region as modified
      if (e->mstart == 0 && e->mend == 0) {
        e->mstart = a->offset;
        e->mend = a->offset + a->count;
      }
      else {
        if (e->mstart > a->offset)
          e->mstart = a->offset;
        if (e->mend < a->offset + a->count)
          e->mend = a->offset + a->count;
      }

      write3res res(NFS3_OK);
      res.resok->count = a->count;
      res.resok->committed = UNSTABLE;
      res.resok->verf = _verf3;
      sbp->reply (&res);
    }
    else {
      // let the cfsrw module write the data to dhash and reply to the
      // in-kernel nfs client
      nsyncwrite++;
      _view->version (wrap(this, &cfsrwcd::dispatch_view, sbp));
    }
  }

  // check if there are any CLOSE or COMMIT that we blocked
  run_rpcs (e);
}

void
cfsrwcd::truncate_cache (nfscall *sbp, file_cache *e, uint64 size)
{
  if (e->afh == 0) {
    str fn = fh2fn(e->fh);
    cfsrwcd::a->open (fn, O_RDWR | O_CREAT, 0666,
	              wrap (this, &cfsrwcd::truncate_cache_open, sbp, e, size));
  }
  else
    truncate_cache_open (sbp, e, size, e->afh, 0);
}

void
cfsrwcd::truncate_cache_open (nfscall *sbp, file_cache *e, uint64 size,
                              ptr<aiofh> afh, int err)
{
  if (!afh) {
    warn << "truncate_cache: open failed: " << err << "\n";
    sbp->reject (SYSTEM_ERR);
    return;
  }
  e->afh = afh;
  assert (e->is_dirty() || e->is_idle() || e->is_open ());

  e->afh->ftrunc (size,
                  wrap (this, &cfsrwcd::truncate_cache_truncate, sbp, e, size));
}

void
cfsrwcd::truncate_cache_truncate (nfscall *sbp, file_cache *e, uint64 size,
                                  int err)
{
  if (err) {
    warn << "truncate_cache: truncate failed: " << err << "\n";
    sbp->reject (SYSTEM_ERR);
    return;
  }
  e->fa.size = size;
  _view->version (wrap (this, &cfsrwcd::dispatch_view, sbp));
}

void
cfsrwcd::flush_done (file_cache *e, nfscall *sbp, bool ok)
{
  assert(e && e->is_flush());
  if (ok) {
    if (e->mstart == e->mend && e->mstart == 0) {
      if (sbp->proc () == cl_NFSPROC3_CLOSE)
        e->open ();
      else
        e->idle ();
    }
    else
      e->dirty ();
    
    commit3res res (NFS3_OK);
    res.resok->verf = _verf3;
    sbp->reply (&res);
  }
  else {
    e->error ();
    sbp->reject (SYSTEM_ERR);
  }

  // invalidate attribute cache
  _ac.attr_enter (e->fh, 0, 0); 

  run_rpcs (e);
}

static inline bool
mtime_recent (nfstime3 mtime)
{
  time_t t = tsnow.tv_sec - mtime.seconds;
  if (t > CACHE_RECENT)
    return false;
  return true;
}

bool
cfsrwcd::dont_run_rpc (nfscall *nc)
{
  if (nc->proc () != NFSPROC3_SETATTR &&
      nc->proc () != NFSPROC3_READ && 
      nc->proc () != NFSPROC3_WRITE &&
      nc->proc () != cl_NFSPROC3_CLOSE &&
      nc->proc () != NFSPROC3_COMMIT)
    return false; 
    
  nfs_fh3 *fh = static_cast<nfs_fh3 *> (nc->getvoidarg ());
  file_cache *e = file_cache_lookup(*fh);
 
  if (!e)
    return false;

  if (e->is_error ()) {
    nc->reject (SYSTEM_ERR);
    return true;
  }
  
  if ((nc->proc () == cl_NFSPROC3_CLOSE || nc->proc () == NFSPROC3_COMMIT)) {
    if (!e->is_dirty () && !e->is_flush () && !e->being_modified ())
      return false;
    else if (e->being_modified ()) {
      WARN(2) << "RPC " << nc->proc () << " blocked due to unfinished writes\n";
      e->rpcs.push_back(nc);
      return true;
    }
  }
    
  // flush mode: block COMMIT, CLOSE, WRITEs and SETATTRs
  if (e->is_flush() && nc->proc () != NFSPROC3_READ) {
    WARN(2) << "RPC " << nc->proc () << " blocked due to flush\n";
    e->rpcs.push_back(nc);
    return true;
  }

  // open mode: on READ and WRITE, fetch file
  else if (e->is_open ()) {
    if (nc->proc () == NFSPROC3_READ || nc->proc () == NFSPROC3_WRITE) {
      // need to re-fetch the file if someone else (not uid(sbp)) has
      // modified the file system, or if uid(sbp) have modified the
      // file but did so a while ago (CACHE_RECENT seconds)

      const fattr3 *f = _ac.last_known_attr_lookup (e->fh);
      vervec v = _ac.current_version ();
      chordID id;
      bool rw = uid (nc, id);
      if (f && 
	  (e->version.equals (v) || 
	   (rw &&
            e->version.equals_except (v, id) &&
	    mtime_recent (e->fa.mtime)))) {
	e->fa = *f;
	e->version = v;
	e->idle ();
	WARN(1) << "fetch skipped " << e->fh << "\n";
      }
      else {
        e->fetch ();
        fill_cache (e, wrap(this, &cfsrwcd::fill_done, e), nc);
	e->rpcs.push_back(nc);
	return true;
      }
    }
  }

  // fetch mode: block WRITEs and SETATTRs. block READ RPCs that
  // cannot be answered.
  else if (e->is_fetch()) {
    if (nc->proc () == NFSPROC3_READ) {
      read3args *a = nc->template getarg<read3args> ();
      uint64 offset = a->offset;
      uint64 size = a->count;
      if (e->received (offset, size))
        return false;
      // *16 forces reading 16 blocks before starting at the
      // beginning again
      offset = (int)(offset/8192) * 8192;
      e->want (offset, size*16, 8192);
    }
    else
      WARN(2) << "RPC " << nc->proc () << " blocked\n";
    e->rpcs.push_back (nc);
    return true;
  }

  return false;
}

bool
cfsrwcd::fcache_dispatch (nfscall *sbp)
{
  if (dont_run_rpc (sbp))
    return true;

  switch (sbp->proc ()) {
    case NFSPROC3_GETATTR:
    {
      file_cache *e = file_cache_lookup (*sbp->template getarg<nfs_fh3> ());
      if (e && (e->is_dirty () || e->is_flush ())) {
        WARN(2) << "getattr reply from cache\n";
        getattr3res res (NFS3_OK);
        *res.attributes = e->fa;
        sbp->reply (&res);
        return true;
      }
    }
    break;

    case NFSPROC3_WRITE:
    {
      write3args *a = sbp->template getarg<write3args> ();
      file_cache *e = file_cache_lookup (a->file);
      if (e) {
        write_to_cache (sbp, e);
        return true;
      }
      else
        warn << "dangling write: " << a->file << "\n";
    }
    break;
  
    case NFSPROC3_READ:
    {
      read3args *a = sbp->template getarg<read3args> ();
      file_cache *e = file_cache_lookup (a->file);
      if (e) {
        read_from_cache (sbp, e);
        return true;
      }
      else
        warn << "dangling read: " << a->file << "\n";
    }
    break;
  
    case cl_NFSPROC3_CLOSE:
    case NFSPROC3_COMMIT:
    {
      nfs_fh3 *a = sbp->template getarg<nfs_fh3> ();
      file_cache *e = file_cache_lookup (*a);
      if (!e) {
        warn << "RPC received without cache entry\n";
	commit3res res (NFS3ERR_STALE);
	sbp->reply (&res);
	return true;
      }
      assert (!e->is_flush ());

      if (e && e->is_dirty ()) {
        flush_cache (e, sbp);
	ncacheflush++;
	return true;
      }

      if (sbp->proc () == cl_NFSPROC3_CLOSE && e &&
	  e->afh != 0 && !e->is_fetch ()) {
        e->afh->close (wrap (&cfsrwcd::file_closed));
        e->afh = 0;
      }

      if (!e)
	warn << "dangling close/commit: " << *a << "\n";

      commit3res res (NFS3_OK);
      res.resok->verf = _verf3;
      sbp->reply (&res);
      return true;
    }
    break;
  
    case NFSPROC3_SETATTR:
    {
      setattr3args *a = sbp->template getarg<setattr3args> ();
      file_cache *e = file_cache_lookup (a->object);
      if (e && !e->is_open ()) {
	apply_sattr (e->fa, a->new_attributes);
        if (a->new_attributes.size.set) {
          truncate_cache (sbp, e, *(a->new_attributes.size.val));
	  return true;
        }
      }
    }
    break;

  default:
    break;
  }

  return false;
}

void
cfsrwcd::fcache_check_reply (nfscall *sbp, void *res)
{
  if (sbp->proc () == NFSPROC3_ACCESS) {
    access3res *ares = static_cast<access3res *> (res);
    if (ares->status == NFS3_OK &&
	ares->resok->obj_attributes.present) {
      access3args *a = sbp->template getarg<access3args> ();
      fattr3 *f = reinterpret_cast<fattr3 *>
        (ares->resok->obj_attributes.attributes.addr ());
      fattr3 fa = *f;
      if (fa.type == NF3REG) {
        file_cache *e = file_cache_lookup(a->object);
        if (!e) {
          file_cache_insert (a->object);
          e = file_cache_lookup(a->object);
          assert(e);
          e->open ();
        }
        else if (e->is_idle())
          e->open ();
      }
    }
  }
}

void
cfsrwcd::file_cache_gc_remove (file_cache *e)
{
  if ((e->is_idle () || e->is_error ()) && e->afh == 0) {
    str fn = fh2fn (e->fh);
    cfsrwcd::a->unlink (fn, wrap (this, &cfsrwcd::file_cache_gc_remove_cb));
    delete e;
  }
}

void
cfsrwcd::file_cache_gc_remove_cb (int)
{
}

