/*
 *
 * Copyright (C) 2002 Benjie Chen (benjie@lcs.mit.edu)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "record.h"
#include "ranges.h"
#include "fcache.h"
#include "cfsrwcd.h"

struct read_obj {
  static const int PARALLEL_READS = 16;
  typedef callback<void,bool,bool>::ref cb_t;

  file_cache *fe;
  cb_t cb;
  cfsrwcd& srv;
  cfsrw& c;
  bigint file;
  uint64 size;
  cfscallp cfscall_ptr;
  unsigned int outstanding_reads;
  bool errorcb;

  void
  read_reply (uint64 off, uint64 cnt, cfscallp c, cfsrw_status status)
  {
    if (errorcb || status) {
      outstanding_reads--;
      fail ();
      return;
    }

    cfsrw_read_res *res = static_cast<cfsrw_read_res *> (c->res ());

    if (res->count > 0) {
      ptr<aiobuf> buf = cfsrwcd::a->bufalloc (res->count);
      if (!buf) {
        cfsrwcd::a->bufwait
	  (wrap (this, &read_obj::read_reply, off, cnt, c, status));
        return;
      }

      memmove(buf->base (), res->data.base (), res->count);
      if (res->count != cnt)
	warn << "asked for " << off << "+" << cnt
	     << ", got " << res->count << "\n";
      fe->afh->write (off, buf, wrap (this, &read_obj::read_reply_write,
	                              off, cnt));
    }
    else {
      if (res->count != cnt)
	warn << "asked for " << off << "+" << cnt << ", got 0\n";
      read_reply_write (off, cnt, 0, 0, 0);
    }
  }

  void
  read_reply_write (uint64 off, uint64 cnt, ptr<aiobuf> buf,
                    ssize_t sz, int err)
  {
    outstanding_reads--;
    
    if (err) {
      warn << "fill_cache: write failed: " << err << "\n";
      fail ();
      return;
    }

    fe->rcv->add (off, cnt);
    do_read ();
    ok ();
  }

  void cfsrw_read (uint64 off, uint32 cnt) 
  {
    fe->req->add(off, cnt);
    outstanding_reads++;

    cfscallp cc = New refcounted<cfscall>
      (CFSRWPROC_READ, wrap (this, &read_obj::read_reply, off, cnt));
    cfsrw_read_arg *arg = static_cast<cfsrw_read_arg *> (cc->arg ());
    arg->file = file;
    arg->offset = off;
    arg->count = cnt;
    cc->l = cfscall_ptr->l;
    cc->s = cfscall_ptr->s;
    cc->start = cfscall_ptr->start;
    cc->stop = cfscall_ptr->stop;
    c.dispatch (cc);
  }

  bool do_read() {
    while (fe->pri.size()) {
      uint64 off = fe->pri.pop_front();
      if (off < size) {
        unsigned s = size-off;
        s = s > 8192 ? 8192 : s;
	if (!fe->req->filled(off, s)) {
          cfsrw_read (off, s);
	  return false;
	}
      }
    }
    uint64 off;
    uint64 cnt;
    if (fe->req->has_next_gap (0, off, cnt)) {
      if (off < size) {
	if (off+cnt > size)
	  cnt = size - off;
        cnt = cnt > 8192 ? 8192 : cnt;
        if (!fe->req->filled(off, cnt)) {
	  cfsrw_read (off, cnt);
	  return false;
        }
      }
    }
    return true;
  }

  static void file_closed (int) {}

  void fail () {
    if (!errorcb) {
      fe->afh->close (wrap (&read_obj::file_closed));
      fe->afh = 0;
      errorcb = true;
      cb (true,false);
    }
    if (outstanding_reads == 0)
      delete this;
  }

  void ok() {
    if (!errorcb)
      cb (outstanding_reads == 0,true);
    if (outstanding_reads == 0)
      delete this;
  }

  void start_read () {
    bool eof = false;
    for (int i=0; i<PARALLEL_READS && !eof; i++)
      eof = do_read ();
    if (!outstanding_reads) // nothing to do
      ok ();
  }

  void truncate_done (int err) {
    if (err)
      fail ();
    else
      start_read ();
  }

  void get_attr (cfscallp c, cfsrw_status status)
  {
    if (!status) {
      WARN(1) << "fetch cache file " << fe->fh << "\n";
      fe->version = cfscall_ptr->start;

      cfsrw_getattr_res *res = static_cast<cfsrw_getattr_res *> (c->res ());
      ptr<fattr3> fa = get_fattr3 (file, &res->attributes);
      fe->fa = *fa;
      size = fe->fa.size;
      fe->afh->ftrunc (size, wrap (this, &read_obj::truncate_done));
    }
    else
      fail ();
  }

  void file_open (ptr<aiofh> afh, int err) 
  {
    if (err) {
      warn << "fill_cache: open failed: " << err << "\n";
      fail ();
      return;
    }
    fe->afh = afh;

    cfscallp cc = New refcounted<cfscall>
      (CFSRWPROC_GETATTR, wrap (this, &read_obj::get_attr));
    cfsrw_getattr_arg *arg = static_cast<cfsrw_getattr_arg *> (cc->arg ());
    arg->object = file;
    cc->l = cfscall_ptr->l;
    cc->s = cfscall_ptr->s;
    cc->start = cfscall_ptr->start;
    cc->stop = cfscall_ptr->stop;
    c.dispatch (cc);
  }

  read_obj (file_cache *e, str fn, cfsrwcd &srv, cfsrw& c,
            vervec start, read_obj::cb_t cb, nfscall *sbp)
    : fe(e), cb(cb), srv(srv), c(c), outstanding_reads(0), errorcb(false)
  {
    file = nfs2cfsfh (&e->fh);
    cfscall_ptr = srv.get_cfscall (sbp, start);

    if (e->afh == 0) {
      cfsrwcd::a->open (fn, O_CREAT | O_RDWR, 0666,
	                wrap (this, &read_obj::file_open));
      return;
    }
    else
      file_open (e->afh, 0);
  }

  ~read_obj() {}
};

void
cfsrwcd::fill_cache (file_cache *e, read_obj::cb_t cb, nfscall *sbp)
{
  _view->version (wrap (this, &cfsrwcd::fill_cache_with_v, e, cb, sbp));
}

void
cfsrwcd::fill_cache_with_v (file_cache *e, read_obj::cb_t cb, nfscall *sbp,
                            vervec v, cfsrw_status status)
{
  if (status) {
    e->error ();
    sbp->error (get_nfsstat3 (status));
    return;
  }

  vNew read_obj (e, fh2fn(e->fh), *this, _cfsrw, v, cb, sbp);
}

