
#include "debug.h"
#include "rxx.h"
#include "ranges.h"
#include "logiter.h"
#include "cfsrw.h"
#include "snapshot.h"

#define SYMLINK_MAX_LENGTH 1024

cfsrw_states::attr::attr ()
{
  ctime = 0;
  nlink = 0;
  truncated = false;
}

void
cfsrw_states::attr::getfa (cfsrw_fattr& fa) const
{
  fa.size = fa.mode = fa.mtime = fa.uid = fa.gid = 0;
  if (sattr.valid_size)
    fa.size = sattr.size;
  if (sattr.valid_mode)
    fa.mode = sattr.mode;
  if (sattr.valid_mtime)
    fa.mtime = sattr.mtime;
  if (sattr.valid_uid)
    fa.uid = sattr.uid;
  if (sattr.valid_gid)
    fa.gid = sattr.gid;
  fa.nlink = nlink;
  fa.ctime = ctime;
  fa.type = type;
}

void
cfsrw_states::attr::update(cfsrw_sattr &in)
{
  if (in.valid_size && !sattr.valid_size) {
    sattr.size = in.size;
    sattr.valid_size = 1;
  }
  if (in.valid_mode && !sattr.valid_mode) {
    sattr.mode = in.mode;
    sattr.valid_mode = 1;
  }
  if (in.valid_mtime && !sattr.valid_mtime) {
    sattr.mtime = in.mtime;
    sattr.valid_mtime = 1;
  }
  if (in.valid_uid && !sattr.valid_uid) {
    sattr.uid = in.uid;
    sattr.valid_uid = 1;
  }
  if (in.valid_gid && !sattr.valid_gid) {
    sattr.gid = in.gid;
    sattr.valid_gid = 1;
  }
}

cfsrw::cfsrw (ref<view> v)
  : _view (v)
{
  char xxb[20];
  rnd.getbytes (xxb, 20);
  mpz_set_rawmag_be (&_verf, xxb, 20);
}

void
cfsrw::scan_log (cbliter cb, vervec start, vervec stop)
{
  vNew refcounted<logiter> (cb, start, stop);
}


//////////////////////////////////////////////////////////////////////
//
//  GETATTR
//
//  find a file's inode record; there's only one, corresponds with
//  file creation. collect incremental file attributes.
//
//////////////////////////////////////////////////////////////////////

void
cfsrw::find_inode (vervec start, vervec stop,
                   bigint object, ptr<cfsrw_states::attr> a,
		   cfsrw_states::cb_attr cb)
{
  scan_log (wrap (this, &cfsrw::find_inode_log_iterator, object, cb, a),
            start, stop);
}

#define update_ctime(X,Y) \
  if(((X)->ctime < (Y)->timestamp ())) { \
    (X)->ctime = (Y)->timestamp (); \
  }

#define update_mtime(X,Y) \
  if(!(X)->sattr.valid_mtime || \
      ((X)->sattr.mtime < (Y)->timestamp ())) { \
    (X)->sattr.mtime = (Y)->timestamp (); \
    (X)->sattr.valid_mtime = 1; \
  }

void
cfsrw::find_inode_log_iterator (bigint object, cfsrw_states::cb_attr cb,
                                ptr<cfsrw_states::attr> attr, iterinfo *i)
{
  if (i->finished) {
    if (i->status)
      cb (attr, i->status);
    else
      cb (attr, CFSRWERR_ABORTED);
    return;
  }
  recordp e = i->record;

  inode_log_record *ie = 0;
  if (e->template record<attr_log_record> () ||
      e->template record<inode_log_record> ()) {
    attr_log_record *ae;
    cfsrw_sattr *sa = 0;
    bigint record_fh = 0;

    if ((ae = e->template record<attr_log_record> ())) {
      sa = ae->attr ();
      record_fh = ae->object ();
    }
    else if ((ie = e->template record<inode_log_record> ())) {
      sa = ie->attr ();
      record_fh = ie->object ();
    }

    if (sa && record_fh == object) {
      /*
       * Use size attribute if it's the most recent in the log AND it
       * sets the file size LARGER or EQUAL than implied by any
       * subsequent writes.  Otherwise ignore it.
       */
      if (sa->valid_size) {
	if (!attr->sattr.valid_size ||
	    (!attr->truncated && sa->size >= attr->sattr.size)) {
          attr->sattr.size = sa->size;
          attr->sattr.valid_size = 1;
        }
        update_mtime (attr, e);
        attr->truncated = true;
      }

      // Make sure file creation sets mtime.
      if (e->template record<inode_log_record>())
        update_mtime (attr, e);

      // Any change to inode sets ctime.
      update_ctime (attr, e);

      // transfer other attributes
      attr->update (*sa);
    }
  }

  // infer size from write_log_record. writes to cfsrw_states::attr
  // don't do write is this file has been truncated
  write_log_record *we = e->template record<write_log_record> ();
  if (we && we->file () == object) {
    // if previously truncated, then don't change the size
    if (!attr->sattr.valid_size || 
	(we->offset() + we->count() > attr->sattr.size && !attr->truncated)) {
      attr->sattr.size = we->offset() + we->count();
      attr->sattr.valid_size = 1;
      update_ctime (attr, we);
    }
    update_mtime (attr, we);
  }

  unlink_log_record *ue = e->template record<unlink_log_record> ();
  if (ue) {
    if (ue->dir() == object) {
      // only effect on directory is to change times.
      update_mtime (attr, ue);
      update_ctime (attr, ue);
      // remove a directory record that is also a directory
      if (ue->object_type () == CFSRW_DIR)
	// unlink from a directory, decr link count
	attr->nlink--;
    }
    else if (ue->object () == object) {
      // always change file's ctime.
      // never change file's mtime.
      // change file's link count.
      update_ctime (attr, ue);
      attr->nlink--;
    }
  }

  rename_log_record *re = e->template record<rename_log_record> ();
  if (re) {
    if (re->dir () == object || re->from_dir () == object) {
      // only effect on directory is to change times.
      update_mtime (attr, re);
      update_ctime (attr, re);
      // add/remove a directory record that is also a directory
      if (re->object_type () == CFSRW_DIR) {
	// rename into a directory, incr link count
	if (re->dir () == object)
          attr->nlink++;
	// rename, dealt with incr link count earlier, here checks if
	// removed a link from same directory
	if (re->from_dir () == object)
	  attr->nlink--;
      }
    }
    else if (re->object () == object) {
      // always change file's ctime.
      // never change file's mtime.
      // don't change file's link count.
      update_ctime (attr, re);
    }
    else if (re->target_exists () && re->target () == object) {
      // rename removed a file
      update_ctime (attr, re);
      attr->nlink--;
    }
  }

  link_log_record *le = e->template record<link_log_record> ();
  if (le) {
    if (le->dir () == object) {
      // only effect on directory is to change times.
      update_mtime (attr, le);
      update_ctime (attr, le);
      // add a directory record that is also a directory
      if (le->object_type () == CFSRW_DIR)
	// link into a directory, incr link count
        attr->nlink++;
    }
    else if (le->object () == object) {
      // always change file's ctime.
      // never change file's mtime.
      // change file's link count.
      update_ctime (attr, le);
      attr->nlink++;
    }
  }

  if (e->template record<inode_log_record> () && ie->object () == object) {
    if (object == _view->root ())
      // hit root, no link record, but we should incr nlink
      attr->nlink++;
    attr->type = ie->object_type ();
    cb (attr, CFSRW_OK);
    i->finished = true;
  }
}

void
cfsrw::found_inode (cfscallp c, bool checked_snapshot,
                    ptr<cfsrw_states::attr> attr, cfsrw_status status)
{
  if (status == CFSRWERR_ABORTED && c->s && !checked_snapshot) {
    cfsrw_getattr_arg *a = static_cast<cfsrw_getattr_arg *>(c->arg ());
    snapshot_operation::find_inode (c->s, a->object, attr,
	                            wrap (this, &cfsrw::found_inode,
				          c, true));
    return;
  }

  if (!status) {
    cfsrw_getattr_res *res = static_cast<cfsrw_getattr_res *>(c->res ());
    attr->getfa (res->attributes);
    // if type is DIR and link count is 0, don't increment it for "."
    // since it has been removed already.  otherwise, we increment
    // nlink by 1.
    if (res->attributes.type == CFSRW_DIR && res->attributes.nlink > 0)
      res->attributes.nlink += 1; // for . (.. doesn't refer to this i-number)
  }

  c->reply (status);
}


//////////////////////////////////////////////////////////////////////
//
//  GETPARENT
//
//  lookup a file or directory by its handle, when successful returns
//  either a rename or link log record that links the file or
//  directory.
//
//////////////////////////////////////////////////////////////////////

void
cfsrw::find_parent (vervec start, vervec stop, bigint object,
                    cfsrw_states::cb_fh cb)
{
  scan_log (wrap (this, &cfsrw::find_parent_log_iterator, object, cb),
                  start, stop);
}

void
cfsrw::find_parent_log_iterator (bigint object, cfsrw_states::cb_fh cb,
                                 iterinfo *i)
{
  cfsrw_status status = CFSRW_OK;
  recordp e;
  unlink_log_record *ur = 0;
  rename_log_record *rr = 0;
  inode_log_record *nr = 0;
  
  e = i->record;
  if (e)
    nr = e->template record<inode_log_record> ();

  if (i->finished || !e || (nr && nr->object() == object)) {
    if (nr && nr->object() == object)
      status = CFSRWERR_NOLINK;
    else
      status = CFSRWERR_ABORTED;
    goto done;
  }
 
  // if object has been removed, return NOENT. do this before we check
  // if object() == object, otherwise target() == object will never be
  // checked
  ur = e->template record<unlink_log_record> ();
  rr = e->template record<rename_log_record> ();
  if ((ur && ur->object() == object)
      || (rr && rr->target_exists () && rr->target() == object)) {
    status = CFSRWERR_NOENT;
    goto done;
  }
  
  // if not a dirop or dirop is not on the object in question, return
  if (!e->dirop_record() || e->dirop_record()->object() != object)
    return;

  WARN(1) << "found object: " << object << " in dir "
          << e->dirop_record()->dir() << "\n";
  cb (e->dirop_record()->dir(), status);
  i->finished = true;
  return;

done:
  WARN(1) << "did not find object " << object << "\n"; 
  cb (0, status);
  i->finished = true;
  return;
}

void
cfsrw::found_parent (cfscallp c, bool checked_snapshot,
                     bigint parent, cfsrw_status status)
{
  if (status == CFSRWERR_ABORTED && c->s && !checked_snapshot) {
    cfsrw_getattr_arg *a = static_cast<cfsrw_getattr_arg *>(c->arg ());
    snapshot_operation::find_parent (c->s, a->object,
	                             wrap (this, &cfsrw::found_parent,
				           c, true));
    return;
  }

  cfsrw_getparent_res *res = static_cast<cfsrw_getparent_res *>(c->res ());
  res->object = parent;
  c->reply (status);
}


//////////////////////////////////////////////////////////////////////
//
//  LOOKUP
//
//  lookup a file name in a directory, given the directory handle.
//
//////////////////////////////////////////////////////////////////////

void
cfsrw::lookup (vervec start, vervec stop,
               bigint dir, str name,
	       cfsrw_states::cb_lookup cb, bool check_intend)
{
  ptr<vec<bigint> > vv = 0;
  if (check_intend)
    vv = New refcounted<vec<bigint> >;
  ref<cfsrw_states::lookup> lu =
    New refcounted<cfsrw_states::lookup> (dir, name, vv);
  scan_log (wrap (this, &cfsrw::lookup_log_iterator, cb, lu), start, stop);
}

void
cfsrw::lookup_log_iterator (cfsrw_states::cb_lookup cb,
                            ref<cfsrw_states::lookup> lu, iterinfo *i)
{
  cfsrw_status status = CFSRW_OK;
  bigint dir = lu->dir;
  str name = lu->name;
  ptr<vec<bigint> > ci = lu->ci;
  
  recordp e;
  
  link_log_record *lr = 0;
  unlink_log_record *ur = 0;
  rename_log_record *rr = 0;
  intend_log_record *ir = 0;
  inode_log_record *nr = 0;
  str lname;
  
  e = i->record;
  if (e)
    nr = e->template record<inode_log_record> ();

  if (i->finished || !e || (nr && nr->object () == dir)) {
    if (nr && nr->object () == dir)
      status = CFSRWERR_NOENT;
    else
      status = CFSRWERR_ABORTED;
    goto done;
  }
  
  lr = e->template record<link_log_record> ();
  rr = e->template record<rename_log_record> ();
  ur = e->template record<unlink_log_record> ();
  ir = e->template record<intend_log_record> ();

  if (ci && ir) {
    // if we run into a non-canceled intend record that matches, stop
    // scan and trigger callback
    //
    // if we run into a canceled intend record, we need to keep on
    // going, but remember that we have seen a canceled intend record
    // by this user.
    
    if (ir->dir () == dir && ir->name () == name) {
      if (ir->canceled ()) {
        ci->push_back (ir->head ());
	return;
      }
      else {
	for (unsigned j = 0; j < ci->size (); j++) {
          if ((*ci)[j] == ir->head ())
	    return;
	}
	// found an intend record that was not canceled
        cb (e, status);
        i->finished = true;
	return;
      }
    }
  }

  // if not right directory or a file by that name has already been
  // deleted or renamed, return.

  if (!(lr || ur || rr) ||
       ((lr && lr->dir () != dir) ||
        (ur && ur->dir () != dir) ||
        (rr && rr->dir () != dir && rr->from_dir () != dir)))
    return;

  if ((ur && ur->name () == name) ||
      (rr && rr->from_name () == name && rr->from_dir () == dir)) {
    status = CFSRWERR_NOENT;
    goto done;
  }

  if (ur)
    return;
  else if (lr)
    lname = lr->name ();
  else if (rr)
    lname = rr->name ();
  else
    assert(0);

  if (lname == name) {
    WARN(1) << "found name: " << name << "\n";
    cb (e, status);
    i->finished = true;
    return;
  }
  return;

done:
  WARN(1) << "did not find name " << name << "\n"; 
  cb ((recordp)0, status);
  i->finished = true;
  return;
}

void
cfsrw::lookup_found (cfscallp c, bool checked_snapshot,
                     recordp e, cfsrw_status status)
{
  if (status == CFSRWERR_ABORTED && c->s && !checked_snapshot) {
    cfsrw_lookup_arg *a = static_cast<cfsrw_lookup_arg *>(c->arg ());
    snapshot_operation::lookup (c->s, a->dir, a->name,
	                        wrap (this, &cfsrw::lookup_found, c, true));
    return;
  }

  if (e && !status) {
    cfsrw_lookup_res *res = static_cast<cfsrw_lookup_res *>(c->res ());
    bigint dir = e->dirop_record()->object ();
    res->object = dir;
  }
  c->reply (status);
}


//////////////////////////////////////////////////////////////////////
//
//  READ
//
//////////////////////////////////////////////////////////////////////

void
cfsrw::read (vervec start, vervec stop, ptr<cfsrw_states::read> rs,
             cfsrw_states::cb_read cb)
{
  memset (rs->buf, '\0', rs->count);
  scan_log (wrap (this, &cfsrw::do_read_got_chunk, rs, cb), start, stop);
}

void
cfsrw::sfrxx (ptr<cfsrw_states::read> rs, write_log_record *e,
              uint64 gs, uint64 gl)
{
  // We have three ranges:
  //   Read RPC asked for offset/count.
  //   Write log record has offset/count.
  //   Gap in ranges has offset/count (gs/gl).
  // We want to copy the intersection of those ranges
  // into buf[].
  //
  uint64 start = rs->offset;
  if (e->offset() > start)
    start = e->offset ();
  if (gs > start)
    start = gs;

  uint64 end = rs->offset + rs->count;
  if (e->offset () + e->count () < end)
    end = e->offset () + e->count ();
  if (gs+gl < end)
    end = gs + gl;
  if (end < start)
    return;

  uint64 bi = start - rs->offset; // index into buf[].
  uint64 wi = start - e->offset (); // index into e->data().
  uint64 n = end - start;

  WARN(2) << "sfrxx " << gs << " " << gl << "\n";

  assert (bi <= rs->count && bi + n <= rs->count);
  assert (wi <= e->count () && wi + n <= e->count ());

  memcpy (rs->buf + bi, e->data().cstr() + wi, n);
  return;
}

void
cfsrw::do_read_got_chunk (ptr<cfsrw_states::read> rs,
                          cfsrw_states::cb_read cb, iterinfo *i)
{
  cfsrw_status status = CFSRW_OK;
  recordp e = i->record;
  if (i->finished || e == 0) {
    status = CFSRWERR_ABORTED;
    i->finished = true;
  }
  else if (inode_log_record *ee = e->template record<inode_log_record> ()) {
    if (ee->object () == rs->object)
      i->finished = true;
  }
  else if (write_log_record *ee = e->template record<write_log_record> ()) {
    if (ee->file () == rs->object) {
      // copy just bytes in intersection of read/write/gaps:
      rs->rng->gaps (wrap (this, &cfsrw::sfrxx, rs, ee));
      // mark this range as done.
      rs->rng->add (ee->offset (), ee->count ());
      // check from _offset to _offset+_count+1. the +1 is there so we
      // can tell if we are at the end of the file or not
      if (rs->rng->check (rs->offset, rs->count + 1))
        i->finished = true;
    }
  }
  else if (attr_log_record *ee = e->template record<attr_log_record> ()) {
    if (ee->object () == rs->object) {
      cfsrw_sattr *a = ee->attr ();
      if (a->valid_size) {
	// size in the first SETATTR is remembered, so we won't return
	// data past the size
	if (!rs->valid_size && a->size > rs->rng->end ()) { 
          rs->valid_size = true;
	  rs->size = a->size;
	}
        // only regions inside the a->size can still be read,
        // everything outside are removed by the setattr
	WARN(2) << "do_read_got_chunk: attr size is " << a->size
	        << " request " << rs->offset << "+" << rs->count << "\n";
        if (a->size >= rs->offset && 
            a->size < (rs->offset+rs->count)) {
          rs->rng->add (a->size, rs->offset+rs->count-a->size);
	  WARN(2) << "cfsrw: attr filled buf " << a->size << " " 
		  << rs->offset+rs->count-a->size  << "\n";
	  WARN(2) << "end is now: " << rs->rng->end () << "\n";
	  // check from _offset to _offset+_count+1. the +1 is there
	  // so we can tell if we are at the end of the file or not
          if (rs->rng->check (rs->offset, rs->count + 1))
            i->finished = true;
        }
        else if (a->size < rs->offset) {
          rs->rng->add (rs->offset, rs->count);
          i->finished = true;
        }
      }
    }
  }

  if (i->finished) {
    uint64 c = rs->rng->end () - rs->offset;
    if (c > rs->count)
      c = rs->count;
    bool eof = (rs->offset + c == rs->rng->end ());
    if (rs->valid_size) {
      if (rs->offset > rs->size) {
	c = 0;
	eof = true;
      }
      else if (rs->offset+c > rs->size) {
	c = rs->size-rs->offset;
	eof = true;
      }
    }
    WARN(2) << "read return with " << rs->offset << "+" << c << "\n";
    rs->rcount = c;
    rs->eof = eof;
    cb (rs, status);
    i->finished = true;
  }
}

void
cfsrw::read_done (cfscallp c, bool checked_snapshot,
                  ptr<cfsrw_states::read> rs, cfsrw_status status)
{
  if (status == CFSRWERR_ABORTED && c->s && !checked_snapshot) {
    snapshot_operation::read (c->s, rs,
	                      wrap (this, &cfsrw::read_done, c, true));
    return;
  }

  if (rs->buf && !status) {
    cfsrw_read_res *res = static_cast<cfsrw_read_res *>(c->res ());
    res->count = rs->rcount;
    res->data.setsize (rs->rcount);
    memmove (res->data.base (), rs->buf, rs->rcount);
    res->eof = rs->eof;
  }
  c->reply (status);
}


//////////////////////////////////////////////////////////////////////
//
//  READLINK
//  
//////////////////////////////////////////////////////////////////////

void
cfsrw::readlink_done (cfscallp c, bool checked_snapshot,
                      ptr<cfsrw_states::read> rs, cfsrw_status status)
{
  if (status == CFSRWERR_ABORTED && c->s && !checked_snapshot) {
    snapshot_operation::read (c->s, rs,
	                      wrap (this, &cfsrw::readlink_done, c, true));
    return;
  }

  if (rs->buf && !status) {
    cfsrw_readlink_res *res = static_cast<cfsrw_readlink_res *>(c->res ());
    str path (rs->buf, rs->rcount);
    res->path = path;
  }
  c->reply (status);
}


//////////////////////////////////////////////////////////////////////
//
//  READDIR
//  
//////////////////////////////////////////////////////////////////////

void
cfsrw::readdir (vervec start, vervec stop,
                bigint dir, ptr<cfsrw_states::readdir> rs,
		cfsrw_states::cb_readdir cb)
{
  scan_log (wrap (this, &cfsrw::scan_dir_log_iterator, dir, rs, cb),
            start, stop);
}

bool
cfsrw::readdir_check_dir_entry (ptr<cfsrw_states::readdir> rs, bigint dir,
                                cfsrw_states::cb_readdir cb,
				str name, bigint object, bool finished)
{
  assert (rs);
  if (name == "") {
    if (finished) {
      rs->res->reply.eof = true; 
      cb (rs, CFSRW_OK);
    }
    else
      cb (rs, CFSRWERR_ABORTED);
    return true;
  }

  // skip entries returned earlier. this won't work for large dir and
  // there are changes between two readdir requests, but then the
  // inconsistency is only for those two requests only, subsequent
  // requests should be fine.
  if (rs->cookie && rs->n < rs->cookie) {
    rs->n++;
    return false;
  }

  // make sure we don't take deleted or renamed inodes into account
  for (unsigned i=0; i<rs->changed.size (); i++) {
    unlink_log_record *ue =
      rs->changed[i]->template record<unlink_log_record> ();
    rename_log_record *re =
      rs->changed[i]->template record<rename_log_record> ();
    if ((ue && ue->name () == name) ||
       // skip if
       // 1. rename/move of some sort affected file in this directory
       // 2. when this name was renamed to something else
       (re &&
          ((re->from_dir () == dir && re->from_name () == name) ||
           (re->target_exists () && re->target () == object))))
      return false;
  }

  if (rs->count < (name.len ()+30)) {
    cb (rs, CFSRW_OK);
    return true;
  }
  else
    rs->count -= (name.len ()+30);

  (*rs->direntp).alloc ();
  (*rs->direntp)->name = name;
  (*rs->direntp)->fileid = cfsfh2fid (object);
  rs->n++;
  (*rs->direntp)->cookie = rs->n;
  rs->direntp = &(*rs->direntp)->nextentry;
  return false;
}

void
cfsrw::scan_dir_log_iterator (bigint dir, ptr<cfsrw_states::readdir> rs,
                              cfsrw_states::cb_readdir cb, iterinfo *i)
{
  recordp e = i->record;
  inode_log_record *n = 0;
  if (e)
    n = e->template record<inode_log_record> ();

  if (i->finished || !e || (n && n->object () == dir)) {
    bool finished = false;
    if (n && n->object () == dir)
      finished = true;
    readdir_check_dir_entry (rs, dir, cb, "", 0, finished);
    i->finished = true;
    return;
  }

  link_log_record *lr = e->template record<link_log_record> ();
  unlink_log_record *ur = e->template record<unlink_log_record> ();
  rename_log_record *rr = e->template record<rename_log_record> ();

  if (!(lr || ur || rr) ||
       ((lr && lr->dir () != dir) ||
        (ur && ur->dir () != dir) ||
        (rr && rr->dir () != dir && rr->from_dir () != dir)))
    return;

  // file deleted or renamed
  if (ur || rr)
    rs->changed.push_back (e);

  // if deleted, bail out (optimization only).
  // if renamed, still consider new name
  if (ur)
    return;

  // found a directory record?
  str name = e->dirop_record()->name ();
  bigint object = e->dirop_record()->object ();
  if (readdir_check_dir_entry (rs, dir, cb, name, object, false)) {
    i->finished = true;
    WARN(1) << "scan_dir_log_iterator found directory record\n"; 
  }
}

void
cfsrw::readdir_done (cfscallp c, bool checked_snapshot,
                     ptr<cfsrw_states::readdir> rs, cfsrw_status status)
{
  if (status == CFSRWERR_ABORTED && c->s && !checked_snapshot) {
    cfsrw_readdir_arg *a = static_cast<cfsrw_readdir_arg *> (c->arg ());
    snapshot_operation::readdir (c->s, a->dir, rs,
	                         wrap (this, &cfsrw::readdir_done, c, true));
    return;
  }

  cfsrw_readdir_res *res = static_cast<cfsrw_readdir_res *> (c->res ());
  res->entries = rs->n;
  c->reply (status);
}


//////////////////////////////////////////////////////////////////////
//
//  SETATTR,UNLINK,WRITE,LINK,RENAME,EXCLDIROP
//
//////////////////////////////////////////////////////////////////////

void
cfsrw::nores_reply (cfscallp c, bool ok)
{
  if (ok)
    c->reply (CFSRW_OK);
  else
    c->reply (CFSRWERR_IO);
}


//////////////////////////////////////////////////////////////////////
//
//  CREATE, SYMLINK, MKDIR
//
//////////////////////////////////////////////////////////////////////

void
cfsrw::create_reply (cfscallp c, ptr<unsigned> status,
                     ptr<int> n, bool ok)
{
  if (ok && *n == 0)
    *status = CFSRW_OK;
  else if (!ok)
    *status = CFSRWERR_IO;
  (*n)++;
  if (*n == 2)
    c->reply ((cfsrw_status)*status);
}

void
cfsrw::symlink_reply (cfscallp c, ptr<unsigned> status,
                      ptr<int> n, bool ok)
{
  if (ok && *n == 0)
    *status = CFSRW_OK;
  else if (!ok)
    *status = CFSRWERR_IO;
  (*n)++;
  if (*n == 3)
    c->reply ((cfsrw_status)*status);
}


//////////////////////////////////////////////////////////////////////
//
//  EXCLDIROP
//
//////////////////////////////////////////////////////////////////////

bool
cfsrw::test_and_set_excldirop (bigint dir, str name)
{
  strbuf x;
  sid (x, dir);
  x << name;
  str xx = x;
  bigint id = compute_hash (xx.cstr(), xx.len());
  bool *hasit = _locks[id];
  if (hasit && *hasit)
    return true;
  _locks.insert (id, true);
  return false;
}

void
cfsrw::clear_excldirop (bigint dir, str name)
{
  strbuf x;
  sid (x, dir);
  x << name;
  str xx = x; 
  bigint id = compute_hash (xx.cstr(), xx.len());
  _locks.remove (id);
}

void
cfsrw::cancel_zombie_intend (cfscallp c, bool again, bool ok)
{
  cfsrw_excldirop_arg *a = static_cast<cfsrw_excldirop_arg *>(c->arg ());
  if (!ok) {
    clear_excldirop (a->dir, a->name);
    c->reply (CFSRWERR_IO);
  }
  else
    lookup (c->start, c->stop, a->dir, a->name,
	    wrap (this, &cfsrw::excldirop_found, c, false, again), true);
}

void
cfsrw::excldirop_cancel (cfscallp c, cfsrw_status s, bool ok)
{
  cfsrw_excldirop_arg *a = static_cast<cfsrw_excldirop_arg *>(c->arg ());
  clear_excldirop (a->dir, a->name);
  if (!ok)
    c->reply (CFSRWERR_IO);
  else
    c->reply (s);
}

void
cfsrw::excldirop_intend (cfscallp c, bool ok)
{
  cfsrw_excldirop_arg *a = static_cast<cfsrw_excldirop_arg *>(c->arg ());
  if (!ok) {
    clear_excldirop (a->dir, a->name);
    c->reply (CFSRWERR_IO);
  }
  else
    _view->version (wrap (this, &cfsrw::excldirop_fetch, c));
}

void
cfsrw::excldirop_fetch (cfscallp c, vervec start, cfsrw_status s)
{
  cfsrw_excldirop_arg *a = static_cast<cfsrw_excldirop_arg *>(c->arg ());
  if (s) {
    clear_excldirop (a->dir, a->name);
    c->reply (CFSRWERR_IO);
    return;
  }
  c->start = start;
  lookup (c->start, c->stop, a->dir, a->name,
          wrap (this, &cfsrw::excldirop_found, c, false, true), true);
}

void
cfsrw::excldirop_found (cfscallp c, bool checked_snapshot, bool again,
                        recordp e, cfsrw_status status)
{
  cfsrw_excldirop_arg *a = static_cast<cfsrw_excldirop_arg *>(c->arg ());

  if (status == CFSRWERR_ABORTED && c->s && !checked_snapshot) {
    snapshot_operation::lookup (c->s, a->dir, a->name,
	                        wrap (this, &cfsrw::excldirop_found,
				      c, true, again));
    return;
  }

  // first, check if an intend record was found
  intend_log_record *intend;
  if (e && (intend = e->template record<intend_log_record> ())) {
    if (intend->verf () == _verf)
      // same verifier, must be our current intend record
      e = 0;
    else if (intend->head () == c->l->head ()) {
      // found another intend record, by us, with a different
      // verifier. this can mean two things: someone is running using
      // the same key on another host, which cfsrw DOES NOT support,
      // or we crashed from a previous server. in either case, CANCEL
      // this record, and try again. we need to TRY AGAIN because
      // someone else may still be trying to create the file, and her
      // INTEND record is hiding behind this bogus INTEND record, and
      // has not been canceled or fulfilled yet.
      ptr<intend_log_record> cancel_record =
	New refcounted<intend_log_record> (intend->verf (),
	                                   a->dir, a->name, true);
      c->l->append
	(_version, cancel_record,
	 wrap (this, &cfsrw::cancel_zombie_intend, c, again));
      return;
    }
  }

  // found a conflicting intend record or a conflicting dirop
  if (e) {
    intend = e->template record<intend_log_record> ();
    if (again) {
      ptr<intend_log_record> cancel_record =
	New refcounted<intend_log_record> (_verf, a->dir, a->name, true);
      c->l->append
	(_version, cancel_record,
	 wrap (this, &cfsrw::excldirop_cancel, c,
	       intend ? CFSRWERR_EXCLCONT : CFSRWERR_EXISTS));
      return;
    }
    clear_excldirop (a->dir, a->name);
    if (intend)
      c->reply (CFSRWERR_EXCLCONT);
    else
      c->reply (CFSRWERR_EXISTS);
    return;
  }

  if (again)
    c->reply (CFSRW_OK);
  else {
    ptr<intend_log_record> intend_record =
      New refcounted<intend_log_record> (_verf, a->dir, a->name, false);
    c->l->append
      (_version, intend_record, wrap (this, &cfsrw::excldirop_intend, c));
  }
}


//////////////////////////////////////////////////////////////////////
//
//  dispatch
//
//////////////////////////////////////////////////////////////////////

void
cfsrw::dispatch (cfscallp c)
{
  if (c->start.gt (_version))
    _version = c->start;

  cfsrw_proc proc = c->proc ();
  switch (proc) {
    case CFSRWPROC_NULL:
      c->reply (CFSRW_OK);
      break;
    case CFSRWPROC_GETATTR:
      {
        cfsrw_getattr_arg *a = static_cast<cfsrw_getattr_arg *>(c->arg ());
	ptr<cfsrw_states::attr> attr = New refcounted<cfsrw_states::attr>;
        find_inode (c->start, c->stop, a->object, attr,
	            wrap (this, &cfsrw::found_inode, c, false));
      }
      break;
    case CFSRWPROC_GETPARENT:
      {
        cfsrw_getparent_arg *a = static_cast<cfsrw_getparent_arg *>(c->arg ());
        find_parent (c->start, c->stop, a->object,
	             wrap (this, &cfsrw::found_parent, c, false));
      }
      break;
    case CFSRWPROC_LOOKUP:
      {
        cfsrw_lookup_arg *a = static_cast<cfsrw_lookup_arg *>(c->arg ());
        lookup (c->start, c->stop, a->dir, a->name,
	        wrap (this, &cfsrw::lookup_found, c, false));
      }
      break;
    case CFSRWPROC_READ:
      {
	cfsrw_read_arg *a = static_cast<cfsrw_read_arg *>(c->arg ());
	ptr<cfsrw_states::read> rs =
	  New refcounted<cfsrw_states::read> (a->file, a->offset, a->count);
	read (c->start, c->stop, rs,
	      wrap (this, &cfsrw::read_done, c, false));
      }
      break;
    case CFSRWPROC_READLINK:
      {
	cfsrw_readlink_arg *a = static_cast<cfsrw_readlink_arg *>(c->arg ());
	ptr<cfsrw_states::read> rs =
	  New refcounted<cfsrw_states::read> (a->file, 0, SYMLINK_MAX_LENGTH);
	read (c->start, c->stop, rs,
	      wrap (this, &cfsrw::readlink_done, c, false));
      }
      break;
    case CFSRWPROC_READDIR:
      {
	cfsrw_readdir_res *r = static_cast<cfsrw_readdir_res *>(c->res ());
	cfsrw_readdir_arg *a = static_cast<cfsrw_readdir_arg *>(c->arg ());
        ptr<cfsrw_states::readdir> rs =
	  New refcounted<cfsrw_states::readdir>(a->cookie, a->count, r);
	readdir (c->start, c->stop, a->dir, rs,
	         wrap (this, &cfsrw::readdir_done, c, false));
      }
      break;
    case CFSRWPROC_SETATTR:
      {
	cfsrw_setattr_arg *a = static_cast<cfsrw_setattr_arg *>(c->arg ());
	recordp new_record =
	  New refcounted<attr_log_record> (a->object, a->new_attributes);
	assert (c->l);
	c->l->append (_version, new_record,
	              wrap (this, &cfsrw::nores_reply, c));
      }
      break;
    case CFSRWPROC_UNLINK:
      {
	cfsrw_unlink_arg *a = static_cast<cfsrw_unlink_arg *>(c->arg ());
        recordp new_record = 
	  New refcounted<unlink_log_record> (a->object, a->type,
	                                     a->dir, a->name);
        assert (c->l);
        c->l->append (_version, new_record,
	              wrap (this, &cfsrw::nores_reply, c));
      }
      break;
    case CFSRWPROC_WRITE:
      {
	cfsrw_write_arg *a = static_cast<cfsrw_write_arg *>(c->arg ());
	recordp new_record =
	  New refcounted<write_log_record> (a->file, a->offset,
	                                    str (a->data.base (), a->count));
	assert (c->l);
	c->l->append (_version, new_record,
	              wrap (this, &cfsrw::nores_reply, c));
      }
      break;
    case CFSRWPROC_CREATE:
    case CFSRWPROC_MKDIR:
      {
	cfsrw_create_arg *a = static_cast<cfsrw_create_arg *>(c->arg ());
	cfsrw_ftype type = (proc == CFSRWPROC_CREATE) ? CFSRW_REG : CFSRW_DIR;
	ptr<inode_log_record> new_inode =
	  New refcounted<inode_log_record> (type, a->attributes);
	ptr<link_log_record> new_link =
	  New refcounted<link_log_record> (new_inode->object (), type,
					   a->dir, a->name);
	
	cfsrw_create_res *res = static_cast<cfsrw_create_res *>(c->res ());
	res->object = new_inode->object ();
	ptr<unsigned> s = New refcounted<unsigned>;
	ptr<int> n = New refcounted<int>;
	*n = 0;
	assert (c->l);
	c->l->append (_version, new_inode,
	              wrap (this, &cfsrw::create_reply, c, s, n));
	c->l->append (_version, new_link,
	              wrap (this, &cfsrw::create_reply, c, s, n));
      }
      break;
    case CFSRWPROC_SYMLINK:
      {
	cfsrw_symlink_arg *a = static_cast<cfsrw_symlink_arg *>(c->arg ());
	
	if (a->linkto.len () > SYMLINK_MAX_LENGTH) {
	  c->reply (CFSRWERR_PATHTOOLONG);
	  return;
	}

	cfsrw_ftype type = CFSRW_LNK;
	ptr<inode_log_record> new_inode =
	  New refcounted<inode_log_record> (type, a->attributes);
	ptr<link_log_record> new_link =
	  New refcounted<link_log_record> (new_inode->object (), type,
					   a->dir, a->name);
	ptr<write_log_record> new_linkto =
	  New refcounted<write_log_record> (new_inode->object (), 0, a->linkto);

	cfsrw_symlink_res *res = static_cast<cfsrw_symlink_res *>(c->res ());
	res->object = new_inode->object ();
	ptr<unsigned> s = New refcounted<unsigned>;
	ptr<int> n = New refcounted<int>;
	*n = 0;
	assert (c->l);
	c->l->append (_version, new_inode,
	              wrap (this, &cfsrw::symlink_reply, c, s, n));
	c->l->append (_version, new_link,
	              wrap (this, &cfsrw::symlink_reply, c, s, n));
	c->l->append (_version, new_linkto,
	              wrap (this, &cfsrw::symlink_reply, c, s, n));
      }
      break;
    case CFSRWPROC_LINK:
      {
	cfsrw_link_arg *a = static_cast<cfsrw_link_arg *>(c->arg ());
	ptr<link_log_record> new_link =
	  New refcounted<link_log_record> (a->object, CFSRW_REG,
					   a->dir, a->name);
	assert (c->l);
	c->l->append (_version, new_link,
	              wrap (this, &cfsrw::nores_reply, c));
      }
      break;
    case CFSRWPROC_RENAME:
      {
	cfsrw_rename_arg *a = static_cast<cfsrw_rename_arg *>(c->arg ());
	recordp new_record;
	if (a->target_exists)
	  new_record = New refcounted<rename_log_record>
	    (a->object, a->type, a->from_dir, a->from_name,
	     a->to_dir, a->to_name, a->target);
	else
	  new_record = New refcounted<rename_log_record>
	    (a->object, a->type, a->from_dir, a->from_name,
	     a->to_dir, a->to_name);
	assert (c->l);
	c->l->append (_version, new_record,
	              wrap (this, &cfsrw::nores_reply, c));
      }
      break;

    case CFSRWPROC_EXCLDIROP:
      {
	cfsrw_excldirop_arg *a = static_cast<cfsrw_excldirop_arg *>(c->arg ());
	if (a->cancel) {
	  ptr<intend_log_record> cancel_record =
	    New refcounted<intend_log_record> (_verf, a->dir, a->name, true);
	  c->l->append
	    (_version, cancel_record,
	     wrap (this, &cfsrw::excldirop_cancel, c, CFSRW_OK));
	}
	else {
	  if (test_and_set_excldirop (a->dir, a->name))
	    c->reply (CFSRWERR_EXCLCONT);
	  else
            lookup (c->start, c->stop, a->dir, a->name,
	            wrap (this, &cfsrw::excldirop_found, c, false, false),
		    true);
	}
      }
      break;
    default:
      fatal ("unknown cfsrw operation\n");
      assert (0);
      break;
  }
}

