
#include "log.h"
#include "cache.h"
#include "serialize.h"
#include "rxx.h"
#include <time.h>

log::log(recordp e, cfsrw_sign_t s)
  : _head(e->head ()), sign(s)
{
  assert(e->template record<head_log_record>());
  _has_snapshot = false;
  _snapshot_new = false;
  _readonly = false;
  _head_record = 0;
  _stores = _failed_stores = 0;
  _refreshed = false;
  _loading = false;
  load_head(e);
}

log::log(chordID head, cfsrw_sign_t s)
  : _head(head), sign(s)
{
  _has_snapshot = false;
  _snapshot_new = false;
  _readonly = false;
  _head_record = 0;
  _stores = _failed_stores = 0;
  _refreshed = false;
  _loading = false;
  locate(_head, true, wrap(mkref(this), &log::load_head));
}

void
log::load_head(recordp e)
{
  assert(!_refreshed);
  if (e) {
    if (_appends_e.size() > 0)
      _refreshed = true;
    _head_record = e;
    head_log_record* h = e->template record<head_log_record>();
    assert(h);
    _root = h->root();
    if (h->has_snapshot() && (!_has_snapshot || !_snapshot_new)) {
      _has_snapshot = true;
      _snapshot_new = false;
      _snapshot = h->snapshot();
    }
  }
}

void
log::load_head_cb(cb_recordp cb, recordp e)
{
  load_head(e);
  cb(e);
}

void
log::append(const vervec& v, recordp e, cbb cb)
{
  assert(!_readonly);

  e->set_version(v);
  _appends_e.push_back(e);
  _appends_cb.push_back(cb);

  if (!_loading) {
    _loading = true;
    // need to load a fresh head record
    locate(_head, true, wrap(mkref(this), &log::load_head_cb,
	                     wrap(mkref(this), &log::do_append)));
  }
  else if (_refreshed)
    do_append(_head_record);
}

void
log::do_append(recordp e)
{
  bool abort = false;
  if (e == 0)
    abort = true;

  for(unsigned i=0; i<_appends_e.size(); i++) {
    recordp e = _appends_e[i];

    if (!abort) {
      if (e->version().conflicts(_head_record->version())) {
        // oops, another writer, in a more up-to-date view perhaps, has
        // written a head of log. that means this view is out of date
        // and should be made read-only!
        abort = true;
        _readonly = true;
        _head_record = 0;
      }
      else {
        vervec v = e->version();
        _head_record->set_version(v);
      }
    }
    _store_cbs.push_back(_appends_cb[i]);

    if (!abort) {
      e->set_prev (_head_record->prev ());
      e->set_head (_head);
      e->set_seqnum (_head_record->seqnum());
      _stores++;
  
      str data = e->serialize();
      chordID id = compute_hash (data.cstr(), data.len());
      _head_record->set_prev (id);
      _head_record->set_seqnum (_head_record->seqnum()+1);
      store_record
        (id, data, 
         wrap(mkref(this), &log::store_head_after_append), e);
    }
  }
  _appends_e.clear();
  _appends_cb.clear();

  if (abort) {
    _stores++;
    store_head_after_append(false);
  }
}


void
log::store_record_cb (cbb cb, dhash_stat status, chordID cid)
{
  if (status != DHASH_OK)
    cb (false);
  else
    cb (true);
}

void
log::store_head (recordp e, const str data, cbb cb,
                 sfs_pubkey2 k, sfs_sig2 s, bool ok)
{
  if (ok)
    cfsrw_dhashclt::insert_pk
      (_head, k, s, data.cstr(), data.len(),
       wrap(mkref(this), &log::store_record_cb, cb), e);
  else {
    store_record_cb(cb, DHASH_ERR, 0);
  }
}

void
log::store_record(chordID key, const str &data, cbb cb, recordp e)
{
  assert(e);
  if (key == _head) {
    assert (sign);
    sign (data, wrap(mkref(this), &log::store_head, e, data, cb));
    return;
  }
  cfsrw_dhashclt::insert_record
    (key, data.cstr(), data.len(),
     wrap (mkref(this), &log::store_record_cb, cb), e);
}

void
log::head_stored(ptr<vec<cbb> >cbs, ptr<head_log_record> h, bool ok)
{
  if (ok && cbs) {
    for (unsigned i=0; i<cbs->size(); i++)
      (*cbs)[i](true);
  }
  else {
    if (cbs) {
      for (unsigned i=0; i<cbs->size(); i++)
        (*cbs)[i](false);
    }
  }
  if (_appends_e.size() > 0) {
    assert(_loading = true);
    _refreshed = true;
    do_append(_head_record);
  }
  else
    _loading = false;
}

void 
log::add_snapshot(chordID s)
{
  _snapshot_new = true;
  _has_snapshot = true;
  _snapshot = s;
}

void
log::store_head_after_append(bool ok)
{
  _stores--;
  if (!ok)
    _failed_stores++;
  if (_stores == 0) {
    if (_failed_stores == 0) {
      _head_record->set_time();
      head_log_record* h = _head_record->template record<head_log_record>();
      if (_has_snapshot) {
	h->snapshot(_snapshot);
	_snapshot_new = false;
      }
      str data = _head_record->serialize();
      ptr<vec<cbb> > cbs = New refcounted<vec<cbb> >();
      for (unsigned i=0; i<_store_cbs.size(); i++)
	cbs->push_back(_store_cbs[i]);
      _store_cbs.clear();
      ptr<head_log_record> e = New refcounted<head_log_record>(*h);
      store_record
	(_head, data, wrap(mkref(this), &log::head_stored, cbs, e), e);
      // leave loading to true and refreshed to false: if a request
      // comes in while DHash is storing the head block, the request
      // is saved, and we won't try to load a new head block until
      // after the old head block is saved.
      _loading = true;
      _refreshed = false;
    }
    else {
      for (unsigned i=0; i<_store_cbs.size(); i++)
	_store_cbs[i](false);
      _store_cbs.clear();
      // start over, a new request would first grab the version of the
      // head block in DHash
      _loading = false;
      _refreshed = false;
    }
    _failed_stores = 0;
  }
}

void
log::locate(chordID id, bool keyhashed, cb_recordp cb)
{
  if (!keyhashed)
    cfsrw_dhashclt::retrieve_record(id, cb);
  else
    cfsrw_dhashclt::retrieve_pk(id, cb);
}

void
log::prev_record(recordp e, cb_recordp cb)
{
  locate(e->prev (), false, cb);
}

void
log::dhash2log(cb_recordp cb, chordID id, ptr<dhash_block> blk)
{
  if (!blk) {
    cb((recordp) 0);
    return;
  }
  recordp e = log_record::parse(str(blk->data, blk->len));
  if (!e) {
    cb((recordp) 0);
    return;
  }
  cb(e);
}

