
#include "serialize.h"
#include "cache.h"
#include "log.h"
#include "logiter.h"
#include "view.h"
#include "sfscd_prot.h"

bool
logiter::check_and_load_record (unsigned i, recordp e)
{
  recordp olde = _entries[i];
  _entries[i] = e;
  _valid[i] = true;

  if (e == 0 || (olde && e->seqnum () >= olde->seqnum ())) {
    _valid[i] = false;
    if (!e) {
      iterinfo info (*this, 0, CFSRWERR_NORECORD, !_comparable, true, true);
      _cb (&info);
      if (info.finished == true)
	return false;
      else
	return true;
    }
    else {
      iterinfo info (*this, e, CFSRWERR_CORRUPTLOG, !_comparable, true, true);
      _cb (&info);
      return false;
    }
  }
  if (olde && !olde->template record<head_log_record> () &&
      (olde->version ()).conflicts (e->version())) {
    iterinfo info (*this, e, CFSRWERR_CONFLICT, !_comparable, true, true);
    _cb (&info);
    return false;
  }
  if (!_stop_at.empty () && e->seqnum () <= _stop_at[i])
    _valid[i] = false;
  return true;
}

void
logiter::check_ordering ()
{
  // there must be an ordering between every two records, otherwise
  // there is either concurrent operations, or conflicts due to stale
  // log head being used
  for (unsigned i=0; i<_entries.size(); i++) {
    if (_valid[i] && !_entries[i]->template record<end_log_record>()) {
      for (unsigned j=0; j<_entries.size(); j++) {
	if (_valid[j] && !_entries[j]->template record<end_log_record>() &&
	    i != j) {
          if ((_entries[i]->version()).equals(_entries[j]->version()) ||
	      (_entries[i]->version()).conflicts(_entries[j]->version())) {
	    _comparable = false;
	    return;
	  }
	}
      }
    }
  }
  _comparable = true;
}

void
logiter::got_version (vervec vv, cfsrw_status status)
{
  if (status) {
    iterinfo i (*this, 0, status, false, true, true);
    _cb (&i);
    return;
  }
  
  if (vv.size () == 0) {
    iterinfo i (*this, 0, CFSRW_OK, false, true, true);
    _cb (&i);
    return;
  }

  _start = vv;
  _nloaded = 0;
  _load_err = 0;
  for (unsigned i=0; i < vv.size (); i++) {
    _valid.push_back (false);
    _entries.push_back ((recordp)0);
    _id.push_back (vv.latest (i));
    if (!_stop.empty()) {
      bool found = false;
      for (unsigned j=0; j < _stop.size() && !found; j++) {
        if (_stop.head(j) == vv.head (i)) {
	  _stop_at.push_back (_stop.seqnum(j));
	  found = true;
        }
      }
      if (!found)
	_stop_at.push_back(0);
    }
    log::locate
      (vv.latest (i), false, wrap(mkref(this), &logiter::load_record, i));
  }
}

void
logiter::load_record (unsigned i, recordp e)
{
  _nloaded++;
  if (!check_and_load_record(i, e)) {
    _load_err = 1;
    return;
  }
  if (_nloaded == _start.size () && !_load_err)
    iter ();
}

void
logiter::got_view (cfsrw_status status)
{
  if (status) {
    iterinfo i (*this, 0, status, false, true, true);
    _cb (&i);
    return;
  }
  _view->version (wrap (mkref(this), &logiter::got_version));
}

logiter::logiter (cbliter cb, vervec vv)
  : _cb(cb)
{
  _comparable = true;
  _start = vv;
  timecb (tsnow, wrap (mkref(this), &logiter::got_version, vv, CFSRW_OK));
}

logiter::logiter (cbliter cb, vervec vv, vervec stop)
  : _cb(cb)
{
  _stop = stop;
  _comparable = true;
  _start = vv;
  timecb (tsnow, wrap (mkref(this), &logiter::got_version, vv, CFSRW_OK));
}

logiter::logiter (cbliter cb, chordID view_addr, vervec stop)
  : _cb(cb)
{
  _stop = stop;
  _comparable = true;
  _view = New refcounted<view>
    (wrap (mkref(this), &logiter::got_view), view_addr);
}

logiter::logiter (cbliter cb, chordID view_addr)
  : _cb(cb)
{
  _comparable = true;
  _view = New refcounted<view>
    (wrap (mkref(this), &logiter::got_view), view_addr);
}

static void
prefetch_record (int n, ptr<log_record> e)
{
  if (n < 4 && e && e->prev () != 0
      && !cfsrw_dhashclt::has_record (e->prev ()))
    log::prev_record (e, wrap(prefetch_record, n+1));
}

void
logiter::iter_cb (unsigned vi, ptr<log_record> e)
{
  if (!check_and_load_record(vi, e))
    return;
  // prefetch log record into record cache
  if (e->prev () != 0 && !cfsrw_dhashclt::has_record (e->prev ()))
    log::prev_record (e, wrap(prefetch_record, 1));
  iter ();
}

void
logiter::iter ()
{
  bool got_hi = false;
  unsigned int hi;

  check_ordering ();
  for (unsigned int i = 0; i < _entries.size (); i++) {
    if (_valid[i] &&
	(got_hi == false ||
	 _entries[i]->version().ge(_entries[hi]->version ()))) {
      hi = i;
      got_hi = true;
    }
  }

  if (got_hi == false) {
    iterinfo i (*this, 0, CFSRW_OK, false, true, true);
    _cb (&i);
    return;
  }
 
  iterinfo i (*this, _entries[hi], CFSRW_OK, !_comparable, false, false);
  _cb (&i);
  if (i.finished)
    return;

  if (i.eol || _entries[hi]->template record<end_log_record> ()) {
    _valid[hi] = false;
    iter ();
  }
  else {
    _id[hi] = _entries[hi]->prev ();
    log::prev_record (_entries[hi], wrap(mkref(this), &logiter::iter_cb, hi));
  }
}

