#ifndef __CFSRW_LOGITER_H
#define __CFSRW_LOGITER_H

#include "dhash.h"
#include "vec.h"
#include "log.h"
#include "view.h"
#include "vervec.h"

class logiter;
class iterinfo;

typedef callback<void, iterinfo *>::ref cbliter;

// logiter combines log records from per-user logs into one virtual
// log. logiter constructors take a callback function, a view id or a
// view_instance object. the callback is called for every log record,
// in sequence, from the most recent to the least recent.
//
// logiter passes to each callback a pointer to an iterinfo object.
// the iterinfo object contains a reference to the logiter object, a
// pointer to the current log record, status, and several flags.
// logiter frees the iterinfo object when the callback returns. thus,
// DO NOT pass around the iterinfo object.
// 
// the finish flag is set to true if the logiter is done. if the
// callback sets the finish flag to be true, upon return from the
// callback, logiter terminates.
//
// the eol flag is set to true if the logiter is done loading and
// processing log records for the log that the current log record is a
// part of. if the callback sets the eol flag to be true, upon return
// from the callback, logiter will no longer load and process log
// records from that log.
//
// the incomparable flag is set to true by logiter if any two of the
// log records it currently keeps in memory have incomparable version
// vectors. logiter keeps one log record from each log at any time.
//
// the status field reports errors. see LI_ERR definitions below. if
// logiter returns LI_ERR_NREC to the callback, the callback can elect
// to contine to scan other logs by setting the finish flag in the
// iterinfo object to false. subsequently, logiter ignores the
// corrupted log. by default, upon LI_ERR_NREC, iterinfo.finished is
// set to true.

class logiter : public virtual refcount { 
private:
  cbliter _cb;
  vervec _start;
  vervec _stop;
  ptr<view> _view;
  unsigned _nloaded;
  int _load_err;

  vec<ptr<log_record> > _entries;
  vec<bool> _valid;
  vec<chordID> _id;
  vec<uint> _stop_at;
  bool _comparable;

  void got_version (vervec vv, cfsrw_status status);
  void got_view (cfsrw_status status);
  void iter_cb (unsigned vi, recordp e);
  void iter ();

  void load_record (unsigned i, recordp e);
  bool check_and_load_record (unsigned i, recordp e);
  void check_ordering ();

public:
  logiter (cbliter cb, vervec vv);
  logiter (cbliter cb, vervec vv, vervec stop);
  logiter (cbliter cb, chordID view_addr);
  logiter (cbliter cb, chordID view_addr, vervec stop);
  ~logiter () {}

  vervec version () const { return _start; }
  unsigned nlogs () const { return _start.size (); }
};

class iterinfo {
public:
  logiter& iter;
  recordp record;
  cfsrw_status status;
  bool incomparable;
  bool finished;
  bool eol;
  iterinfo(logiter &i, recordp e, cfsrw_status s, bool c, bool f, bool l)
    : iter(i), record(e), status(s), incomparable(c), finished(f), eol(l) {}
  iterinfo(const iterinfo& i)
    : iter(i.iter), record(i.record), status(i.status),
      incomparable(i.incomparable), finished(i.finished), eol(i.eol) {}
};

#endif // __CFSRW_LOG_H

