
#include "async.h"  // tsnow
#include "record.h"
#include "serialize.h"
#include <time.h>

void
cfsrw_sattr::serialize(strbuf &sb) const
{
  sint (sb, valid_size);
  squad (sb, size);
  sint (sb, valid_mode);
  sint (sb, mode);
  sint (sb, valid_mtime);
  sint (sb, mtime);
  sint (sb, valid_uid);
  sint (sb, uid);
  sint (sb, valid_gid);
  sint (sb, gid);
}

void
cfsrw_sattr::xinxin(xin *in)
{
  in->rint (valid_size);
  in->rquad (size);
  in->rint (valid_mode);
  in->rint ((int) mode);
  in->rint (valid_mtime);
  in->rint ((int) mtime);
  in->rint (valid_uid);
  in->rint ((int) uid);
  in->rint (valid_gid);
  in->rint ((int) gid);
}

log_record::log_record (int type)
{
  _record_type = type;
  set_time ();
}

void
log_record::set_time ()
{
  struct timeval t;
  gettimeofday (&t, 0);
  _ts = t.tv_sec;
}

ptr<log_record>
log_record::parse(str ins)
{
  ptr<log_record> e = 0;
  xin *in = New xin(ins);
  
  chordID prev;
  chordID head;
  int seqnum;
  int type;
  int sec;
  int vsize;
  vervec version;
 
  if (!in->rid(prev))
    goto bad;
  if (!in->rid(head))
    goto bad;
  if (!in->rint(seqnum))
    goto bad;
  if (!in->rint(type))
    goto bad;
  if (!in->rint(sec))
    goto bad;
  if (!in->rint(vsize))
    goto bad;
  for (int i=0; i<vsize; i++) {
    chordID h;
    chordID l;
    int s;
    if (!in->rid(h))
      goto bad;
    if (!in->rid(l))
      goto bad;
    if (!in->rint(s))
      goto bad;
    version.add(h, l, s);
  }

  if (type == T_END)
    e = New refcounted<end_log_record> (in);
  else if (type == T_INODE)
    e = New refcounted<inode_log_record> (in);
  else if (type == T_HEAD)
    e = New refcounted<head_log_record> (in);
  else if (type == T_LINK)
    e = New refcounted<link_log_record> (in);
  else if (type == T_UNLINK)
    e = New refcounted<unlink_log_record> (in);
  else if (type == T_RENAME)
    e = New refcounted<rename_log_record> (in);
  else if (type == T_WRITE)
    e = New refcounted<write_log_record> (in);
  else if (type == T_INTEND)
    e = New refcounted<intend_log_record> (in);
  else if (type == T_ATTR)
    e = New refcounted<attr_log_record> (in);
  else
    warn << "log_record::parse: unknown record type " << type << "\n";
  delete in;

  if (e) {
    e->_prev = prev;
    e->_head = head;
    e->_record_type = type;
    e->_seqnum = (uint)seqnum;
    e->_ts = sec;
    e->_version = version;
    return e;
  }
  else
    return 0;
    
bad:  
  delete in;
  return 0;
}
 
chordID
log_record::hash () const
{
  str data = serialize ();
  return compute_hash (data.cstr (), data.len ());
}

ptr<log_record>
log_record::dup()
{
  ptr<log_record> r = 0;
  if (record_type () == T_END)
    r = New refcounted<end_log_record>
      (*this->template record<end_log_record> ());
  else if (record_type () == T_INODE)
    r = New refcounted<inode_log_record>
      (*this->template record<inode_log_record> ());
  else if (record_type () == T_HEAD)
    r = New refcounted<head_log_record>
      (*this->template record<head_log_record> ());
  else if (record_type () == T_LINK)
    r = New refcounted<link_log_record>
      (*this->template record<link_log_record> ());
  else if (record_type () == T_UNLINK)
    r = New refcounted<unlink_log_record>
      (*this->template record<unlink_log_record> ());
  else if (record_type () == T_RENAME)
    r = New refcounted<rename_log_record>
      (*this->template record<rename_log_record> ());
  else if (record_type () == T_WRITE)
    r = New refcounted<write_log_record>
      (*this->template record<write_log_record> ());
  else if (record_type () == T_INTEND)
    r = New refcounted<intend_log_record>
      (*this->template record<intend_log_record> ());
  else if (record_type () == T_ATTR)
    r = New refcounted<attr_log_record>
      (*this->template record<attr_log_record> ());
  else
    assert (0);
  return r;
}

void
log_record::serialize1(strbuf &sb) const
{
  sid(sb, _prev);
  sid(sb, _head);
  sint(sb, _seqnum);
  sint(sb, _record_type);
  sint(sb, _ts);
  _version.serialize(sb);
}

attr_log_record::attr_log_record (bigint object, cfsrw_sattr attr)
  : log_record(T_ATTR), _attr(attr)
{
  _object = object;
}

attr_log_record::attr_log_record(xin *in)
{
  in->rid(_object);
  _attr.xinxin(in);
}

str
attr_log_record::serialize() const
{
  strbuf sb;

  serialize1(sb);
  sid(sb, _object);
  _attr.serialize(sb);
  return sb;
}

inode_log_record::inode_log_record (cfsrw_ftype object_type, bigint object)
  : log_record (T_INODE), _attr ()
{
  _object_type = object_type;
  _object = object;
}

inode_log_record::inode_log_record (cfsrw_ftype object_type, bigint object,
                                    cfsrw_sattr sa)
  : log_record (T_INODE), _attr (sa)
{
  _object_type = object_type;
  _object = object;
}

inode_log_record::inode_log_record (cfsrw_ftype object_type)
  : log_record (T_INODE), _attr ()
{
  _object_type = object_type;
  // random new i-number.
  char xxb[20];
  rnd.getbytes(xxb, 20);
  mpz_set_rawmag_be(&_object, xxb, 20);
}

inode_log_record::inode_log_record (cfsrw_ftype object_type, cfsrw_sattr sa)
  : log_record (T_INODE), _attr (sa)
{
  _object_type = object_type;
  // random new i-number.
  char xxb[20];
  rnd.getbytes(xxb, 20);
  mpz_set_rawmag_be(&_object, xxb, 20);
}

inode_log_record::inode_log_record (xin *in)
{
  int i;
  in->rint(i);
  _object_type = (cfsrw_ftype) i;
  in->rid(_object);
  _attr.xinxin(in);
}

str
inode_log_record::serialize () const
{
  strbuf sb;
  serialize1(sb);
  sint(sb, _object_type);
  sid(sb, _object);
  _attr.serialize(sb);
  return sb;
}

end_log_record::end_log_record()
  : log_record(T_END)
{
  set_prev (0);
}

end_log_record::end_log_record(xin *)
{
}

str
end_log_record::serialize() const
{
  strbuf sb;
  serialize1(sb);
  return sb;
}

head_log_record::head_log_record(bigint root)
  : log_record(T_HEAD), _has_snapshot(false), _snapshot (0)
{
  _root = root;
}

head_log_record::head_log_record()
  : log_record(), _has_snapshot(false), _snapshot (0)
{
}

head_log_record::head_log_record(xin *in)
{
  int a;
  in->rid(_root);
  in->rint(a);
  if (a) _has_snapshot = true;
  else _has_snapshot = false;
  in->rid(_snapshot);
}

str
head_log_record::serialize() const
{
  strbuf sb;
  serialize1(sb);
  sid(sb, _root);
  int a = _has_snapshot ? 1 : 0;
  sint(sb, a);
  sid(sb, _snapshot);
  return sb;
}

dirop_log_record::dirop_log_record
  (int record_type, bigint object, cfsrw_ftype object_type,
   bigint dir, str name)
  : log_record (record_type),
    _object (object), _object_type (object_type), _dir(dir), _name(name)
{
}

link_log_record::link_log_record
  (bigint object, cfsrw_ftype object_type, bigint dir, str name)
  : dirop_log_record (T_LINK, object, object_type, dir, name)
{
}

link_log_record::link_log_record(xin *in)
{
  in->rid(_object);
  int i;
  in->rint(i);
  _object_type = (cfsrw_ftype) i;
  in->rid(_dir);
  in->rstr(_name);
}

str
link_log_record::serialize() const
{
  strbuf sb;
  serialize1(sb);
  sid(sb, _object);
  sint(sb, (int) _object_type);
  sid(sb, _dir);
  sstr(sb, _name);
  return sb;
}

unlink_log_record::unlink_log_record
  (bigint object, cfsrw_ftype object_type, bigint dir, str name)
  : dirop_log_record (T_UNLINK, object, object_type, dir, name)
{
}

unlink_log_record::unlink_log_record(xin *in)
{
  in->rid(_object);
  int i;
  in->rint(i);
  _object_type = (cfsrw_ftype) i;
  in->rid(_dir);
  in->rstr(_name);
}

str
unlink_log_record::serialize() const
{
  strbuf sb;
  serialize1(sb);
  sid(sb, _object);
  sint(sb, (int) _object_type);
  sid(sb, _dir);
  sstr(sb, _name);
  return sb;
}

rename_log_record::rename_log_record
  (bigint object, cfsrw_ftype object_type,
   bigint from_dir, str from_name,
   bigint dir, str name, bigint target)

  : dirop_log_record (T_RENAME, object, object_type, dir, name),
    _from_dir (from_dir), _from_name (from_name),
    _target_exists (true), _target (target)
{
}

rename_log_record::rename_log_record
  (bigint object, cfsrw_ftype object_type,
   bigint from_dir, str from_name,
   bigint dir, str name)
  : dirop_log_record (T_RENAME, object, object_type, dir, name),
    _from_dir (from_dir), _from_name (from_name), _target_exists (false)
{
}

rename_log_record::rename_log_record(xin *in)
{
  in->rid(_object);
  int i;
  in->rint(i);
  _object_type = (cfsrw_ftype) i;
  in->rid(_from_dir);
  in->rstr(_from_name);
  in->rid(_dir);
  in->rstr(_name);
  int n;
  in->rint(n);
  if (n)
    _target_exists = true;
  else
    _target_exists = false;
  in->rid(_target);
}

str
rename_log_record::serialize() const
{
  strbuf sb;
  serialize1(sb);
  sid(sb, _object);
  sint(sb, (int) _object_type);
  sid(sb, _from_dir);
  sstr(sb, _from_name);
  sid(sb, _dir);
  sstr(sb, _name);
  if (_target_exists)
    sint(sb, 1);
  else
    sint(sb, 0);
  sid(sb, _target);
  return sb;
}

intend_log_record::intend_log_record(bigint v, bigint d, str name, bool c)
  : log_record(T_INTEND), _verf(v), _dir(d), _name(name)
{
  if (c)
    _canceled = 1;
  else
    _canceled = 0;
}

intend_log_record::intend_log_record(xin *in)
{
  in->rid(_verf);
  in->rid(_dir);
  in->rstr(_name);
  in->rint(_canceled);
}

str
intend_log_record::serialize() const
{
  strbuf sb;
  serialize1(sb);
  sid(sb, _verf);
  sid(sb, _dir);
  sstr(sb, _name);
  sint(sb, _canceled);
  return sb;
}

write_log_record::write_log_record(bigint file, uint64 offset, str data)
  : log_record(T_WRITE), _data(data)
{
  _file = file;
  _offset = offset;
}

write_log_record::write_log_record(xin *in)
{
  in->rid(_file);
  in->rquad(_offset);
  in->rstr(_data);
}

str
write_log_record::serialize() const
{
  strbuf sb;
  serialize1(sb);
  sid(sb, _file);
  squad(sb, _offset);
  sstr(sb, _data);
  return sb;
}

