#ifndef CFSRW_RECORD_H
#define CFSRW_RECORD_H

#include "vec.h"
#include "vervec.h"
#include "serialize.h"
#include "cfsrw_prot.h"

class head_log_record;
class end_log_record;
class dirop_log_record;
class attr_log_record;
class write_log_record;
class inode_log_record;
class intend_log_record;
class link_log_record;
class unlink_log_record;
class rename_log_record;

class log_record {
private:
  chordID _prev;
  chordID _head;
  int _record_type;
  uint _seqnum;
  unsigned _ts;
  vervec _version;

protected:
  void serialize1 (strbuf &sb) const;
  virtual const std::type_info& cxxtype () const { return typeid(*this); }
  
  friend class log;
  void set_time ();

public:
  log_record (int type = 0);
  virtual ~log_record () {}
  ptr<log_record> dup ();
  
  int record_type ()       const { return _record_type; }
  virtual str serialize () const = 0;
  unsigned timestamp ()    const { return _ts; }

  chordID hash ()   const; 
  chordID prev ()   const { return _prev; }
  chordID head ()   const { return _head; }
  uint seqnum ()    const { return _seqnum; }
  vervec version () const { return _version; }
  
  void set_prev (chordID id) { _prev = id; }
  void set_head (chordID id) { _head = id; }
  void set_seqnum (uint n)   { _seqnum = n; }
  void set_version (const vervec &v) { _version = v; }

  enum record_type {
    T_END = 1,
    T_INODE,
    T_HEAD,
    T_LINK,
    T_UNLINK,
    T_WRITE,
    T_ATTR,
    T_RENAME,
    T_INTEND
  };

  template<class T> T* record () {
    if (typeid (T) != cxxtype () &&
	// not sure why cxxtype() can return a refcounted type, since
	// supposedly the this pointer is just a plain log_record ptr.
	typeid (refcounted<T>) != cxxtype ())
      return 0;
    return static_cast<T*> (this);
  }

  dirop_log_record * dirop_record () {
    if (_record_type == T_LINK || _record_type == T_UNLINK ||
	_record_type == T_RENAME)
      return ((dirop_log_record*) this);
    return 0;
  }

  static ptr<log_record> parse (str);
};


class attr_log_record : public log_record
{
private:
  bigint _object;
  cfsrw_sattr _attr;

public:
  attr_log_record (const attr_log_record& r) { *this = r; }
  attr_log_record (bigint, cfsrw_sattr);
  attr_log_record (xin *);

  bigint object () const { return _object; }
  cfsrw_sattr* attr () { return &_attr; }
  str serialize () const;
};


class inode_log_record : public log_record
{
private:
  cfsrw_ftype _object_type;
  bigint _object;
  cfsrw_sattr _attr;

public:
  inode_log_record (const inode_log_record& r) { *this = r; }
  inode_log_record (cfsrw_ftype);
  inode_log_record (cfsrw_ftype, cfsrw_sattr);
  inode_log_record (cfsrw_ftype, bigint);
  inode_log_record (cfsrw_ftype, bigint, cfsrw_sattr);
  inode_log_record (xin *);

  cfsrw_ftype object_type () const { return _object_type; }
  bigint object () const { return _object; }
  cfsrw_sattr* attr () { return &_attr; }
  str serialize () const;
};


class end_log_record : public log_record
{
public:
  end_log_record ();
  end_log_record (const end_log_record& r) { *this = r; }
  end_log_record (xin *);

  str serialize () const;
};


class head_log_record : public log_record
{
private:
  bigint _root;
  bool _has_snapshot;
  chordID _snapshot;
  
public:
  head_log_record ();
  head_log_record (const head_log_record& h) { *this = h; }
  head_log_record (bigint root);
  head_log_record (xin *);

  bigint root () const { return _root; }
  bool has_snapshot () const { return _has_snapshot; }
  chordID snapshot () const { return _snapshot; }
  void snapshot (chordID s) { _snapshot = s; _has_snapshot = true; }
  str serialize() const;
};


class intend_log_record : public log_record
{
private:
  bigint _verf;
  bigint _dir;
  str _name;
  int _canceled;

public:
  intend_log_record (const intend_log_record& r) { *this = r; }
  intend_log_record (bigint, bigint, str, bool);
  intend_log_record (xin *);

  str serialize () const;
  str name () const { return _name; }
  bigint verf () const { return _verf; }
  bigint dir () const { return _dir; }
  bool canceled () const { return (_canceled == 1); }
};


class write_log_record : public log_record
{
private:
  bigint _file;
  uint64 _offset;
  str _data;

public:
  write_log_record (const write_log_record& r) { *this = r; }
  write_log_record (bigint, uint64, str);
  write_log_record (xin *);

  bigint file () const { return _file; }
  uint64 offset () const { return _offset; }
  uint32 count () const { return _data.len(); }
  str data () const { return _data; }
  str serialize () const;
};


class dirop_log_record : public log_record
{
protected:
  bigint _object;
  cfsrw_ftype _object_type;
  bigint _dir;
  str _name;
  
  dirop_log_record (int record_type, bigint object, cfsrw_ftype object_type, 
                    bigint dir, str name);
  dirop_log_record () {}  // used by copy constructors of subclasses

public:
  virtual str serialize () const = 0;
  
  bigint dir () const { return _dir; }
  bigint object () const { return _object; }
  str name () const { return _name; }
  cfsrw_ftype object_type () const { return _object_type; }
};


class link_log_record : public dirop_log_record
{
public:
  link_log_record (const link_log_record &r) { *this = r; }
  link_log_record (bigint object, cfsrw_ftype object_type,
                   bigint dir, str name);
  link_log_record (xin *);
  str serialize () const;
};


class unlink_log_record : public dirop_log_record
{
public:
  unlink_log_record (const unlink_log_record &r) { *this = r; }
  unlink_log_record (bigint object, cfsrw_ftype object_type,
                     bigint dir, str name);
  unlink_log_record (xin *);
  str serialize () const;
};


class rename_log_record : public dirop_log_record
{
private:
  bigint _from_dir;
  str _from_name;
  bool _target_exists;
  bigint _target;

public:
  rename_log_record (const rename_log_record &r) { *this = r; }
  rename_log_record (bigint object, cfsrw_ftype object_type,
                     bigint from_dir, str from_name,
                     bigint dir, str name, bigint target);
  rename_log_record (bigint object, cfsrw_ftype object_type,
                     bigint from_dir, str from_name,
                     bigint dir, str name);
  rename_log_record (xin *);
  str serialize () const;
  
  bigint from_dir ()  const { return _from_dir; }
  str from_name ()     const { return _from_name; }
  bool target_exists ()    const { return _target_exists; }
  bigint target () const { return _target; }
};

typedef ptr<log_record> recordp;
typedef callback<void, recordp>::ref cb_recordp;

inline uint64
cfsfh2fid (bigint object)
{
  char ib[20];
  mpz_get_rawmag_be (ib, sizeof (ib), &object);
  uint64 fileid;
  bcopy (ib, &fileid, sizeof (fileid));
  return fileid;
}

#endif

