
#include "cache.h"
#include "dhash.h"
#include "record.h"
#include "log.h"
#include "view.h"

// view constructor

view_constructor::view_constructor (const view &v)
{
  root (v.root ());
  for (unsigned i = 0; i < v.nlogs (); i++)
    add_log (v.head (i), v.latest (i));
}

chordID
view_constructor::store (callback<void, dhash_stat, chordID>::ref cb)
{
  str data = serialize ();
  chordID id = compute_hash (data.cstr (), data.len ());
  cfsrw_dhashclt::insert_ch (data.cstr (), data.len (), cb);
  return id;
}

void
view_constructor::sort ()
  // slow but simple sort
{
  vec<chordID> h;
  vec<chordID> l;
  vec<bool> done;
  bool found = true;
  for (unsigned i=0; i<_head.size(); i++)
    done.push_back(false);

  while (found) {
    found = false;
    chordID sv;
    unsigned s = 0;
    for (unsigned i=0; i<_head.size(); i++) {
      if (!done[i]) {
	if (!found || _head[i] < sv) {
          found = true;
	  sv = _head[i];
	  s = i;
	}
      }
    }
    if (found) {
      h.push_back(_head[s]);
      l.push_back(_latest[s]);
      done[s] = true;
    }
  }

  _head = h;
  _latest = l;
}

str
view_constructor::serialize ()
{
  sort (); // paranoia
  strbuf sb;
  sid (sb, _root);
  sint (sb, _head.size ());
  for (unsigned i=0; i<_head.size (); i++) {
    sid (sb, _head[i]);
    sid (sb, _latest[i]);
  }
  return sb;
}

void
view_constructor::add_log (chordID h, chordID b)
{
  for (unsigned i=0; i<_head.size (); i++) {
    if (_head[i] == h) {
      _latest[i] = b;
      return;
    }
  }
  _head.push_back (h);
  _latest.push_back (b);
  sort ();
}

void
view_constructor::parse (xin *in)
{
  int n;
  in->rid (_root);
  in->rint (n);
  for (int i=0; i<n; i++) {
    chordID h, b;
    in->rid (h);
    in->rid (b);
    add_log (h, b);
  }
}

// view object

view::view (callback<void, cfsrw_status>::ref cb, chordID addr)
{
  ptr<view_constructor> vc = New refcounted<view_constructor>;
  cfsrw_dhashclt::retrieve_ch (addr, wrap(this, &view::got_view, vc, cb));
}

view::view (ptr<view_constructor> c)
{
  assert (c);
  take (c);
}

void
view::take (ptr<view_constructor> c)
{
  assert (c);
  _root = c->_root;
  for (unsigned i = 0; i < c->_head.size (); i++) {
    _head.push_back (c->_head[i]);
    _latest.push_back (c->_latest[i]);
  }
}

void
view::got_view (ptr<view_constructor> vc, 
                callback<void, cfsrw_status>::ref cb, ptr<dhash_block> blk)
{
  if (blk == 0) {
    warn << "could not fetch view block\n";
    cb (CFSRWERR_NOVIEW);
    return;
  }

  str ins (blk->data, blk->len);
  xin *in = New xin (ins);
  vc->parse (in);
  delete in;
  take (vc);
  cb (CFSRW_OK);
}

void
view::version (callback<void, vervec, cfsrw_status>::ref cb)
{
  ptr<view_state> v = New refcounted<view_state> ();
  for (unsigned i=0; i<_head.size (); i++) {
    v->latest.push_back (_latest[i]);
    v->seqnum.push_back (0);
    if (_head[i] == _latest[i])
      // need to load loghead first
      log::locate
        (_head[i], true, wrap(this, &view::load_record, true, v, i, cb));
    else
      log::locate
        (_latest[i], false, wrap(this, &view::load_record, false, v, i, cb));
  }
  load_done (v, cb);
}

void
view::load_record (bool ishead, ptr<view_state> v, unsigned i,
                   callback<void, vervec, cfsrw_status>::ref cb, recordp e)
{
  if (!e) {
    v->nrecords++;
    v->error = ishead ? CFSRWERR_NOLOGHEAD : CFSRWERR_NORECORD;
  }
  else if (ishead) {
    if (e->template record<head_log_record>()) {
      v->latest[i] = e->prev ();
      log::locate
        (v->latest[i], false, wrap(this, &view::load_record, false, v, i, cb));
    }
    else {
      v->nrecords++;
      v->error = CFSRWERR_NOTLOGHEAD;
    }
  }
  else {
    v->nrecords++;
    v->seqnum[i] = e->seqnum ();
  }
  load_done (v, cb);
}

void
view::load_done (ptr<view_state> v,
                 callback<void, vervec, cfsrw_status>::ref cb)
{
  assert (v->nrecords <= _head.size ());
  if (v->nrecords != _head.size ())
    return;

  vervec vv;
  if (!v->error) {
    for (unsigned i=0; i<_head.size(); i++)
      vv.add (_head[i], v->latest[i], v->seqnum[i]);
  }
  cb (vv, v->error);
}

