
/*
 * Copyright (C) 2002 Athicha Muthitacharoen (athicha@lcs.mit.edu)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "debug.h"
#include "snapshot.h"

#define SNAPSHOT_DBLKSIZE 256 // number of entries per directory block

aiod* snapshot::a = New aiod (2);

int
snapshot::init ()
{
  if (mkdir("/var/tmp/cfsrw-scache", 0755) < 0 && errno != EEXIST) {
    warn << "cannot create directory /var/tmp/cfsrw-scache\n";
    return -1;
  }
  for (unsigned i=0; i<240; i++) {
    strbuf s;
    s << "/var/tmp/cfsrw-scache/";
    s << i;
    if (mkdir(str(s).cstr(), 0755) < 0 && errno != EEXIST) {
      warn << "cannot create directory " << str(s) << "\n";
      return -1;
    }
  }
  return 0;
}

snapshot::snapshot(chordID sid, cb_build cb)
  : prev_link (0), id (0), prev_snapshot (0), vid (0), bcb(cb)
{
  insert_failures = 0;
  outstanding_inserts = 0;
  outstanding_treeops = 0;
  inserts = 0;
  fmap_db = 0;
  tree_build_started = false;
  cfsrw_dhashclt::retrieve_ch
    (sid, wrap (mkref(this), &snapshot::get_info_cb, sid));
}

void
snapshot::get_info_cb(chordID sid, ptr<dhash_block> blk)
{
  if (!blk) {
    warn << "snapshot::get_info_cb: can't find snapshot info block.\n";
    bcb (0, (ptr<snapshot>)0);
    return;
  }
  parse_info (str (blk->data, blk->len));
  cfsrw_dhashclt::retrieve_ch
    (id, wrap (mkref(this), &snapshot::get_fmap_cb, sid));
}

void 
snapshot::get_fmap_cb(chordID sid, ptr<dhash_block> blk) 
{
  if (!blk) {
    warn << "snapshot::get_fmap_cb: can't find snapshot block.\n";
    bcb (0, (ptr<snapshot>)0);
    return;
  }
  parse_fmap (true, str (blk->data, blk->len), sid);
}

void 
snapshot::get_more_fmap_cb(chordID sid, ptr<dhash_block> blk) 
{
  if (!blk) {
    warn << "snapshot::get_more_fmap_cb: can't find snapshot block.\n";
    bcb (0, (ptr<snapshot>)0);
    return;
  }
  parse_fmap (false, str (blk->data, blk->len), sid);
}

void 
snapshot::parse_info (str ins)
{
  xin *in = New xin (ins);
  in->rid (prev_link);
  WARN(2) << "parse_info: prev_link " << prev_link << "\n";
  in->rid (id);
  WARN(2) << "parse_info: id " << id << "\n";
  delete in;
}

void
snapshot::parse_fmap (bool first, str ins, chordID sid)
{
  xin *in = New xin (ins);

  if (first) {
    del (); // clear all vectors first
    int n;
    in->rid (prev_snapshot);
    in->rid (vid);
    in->rid (rootdir);
    in->rint (n);
    for (int i=0; i<n; i++) {
      chordID h, l;
      int s;
      in->rid (h);
      in->rid (l);
      in->rint (s);
      _version.add(h, l, s);
    }
  
    if (!(fmap_db = fmap_createdb ())) {
      warn << "cannot create fmap database.\n";
      bcb (0, (ptr<snapshot>)0);
      return;
    }
    fmap_entry_inserts = 0;
    fmap_entry_insert_failures = 0;
    fmap_load_done = false;
  }

  int n;
  int remain;
  chordID next;

  in->rint (n);
  in->rint (remain);
  in->rid  (next);

  for (int i=0; i<n; i++) {
    chordID inum, chash;
    in->rid (inum);
    in->rid (chash);

    ptr<dbrec> k = id2dbrec (inum);
    ptr<dbrec> d = id2dbrec (chash);
    fmap_entry_inserts++;
    fmap_db->insert
      (k, d, wrap(mkref(this), &snapshot::fmap_entry_load_cb, sid));
  }
  delete in;

  if (remain > 0) {
    WARN(1) << "load extra fmap page at " << next << "\n";
    cfsrw_dhashclt::retrieve_ch
      (next, wrap (mkref(this), &snapshot::get_more_fmap_cb, sid));
  }
  else {
    fmap_load_done = true;
    if (fmap_entry_inserts == 0) {
      if (fmap_entry_insert_failures == 0)
        bcb (sid, mkref(this));
      else
        bcb (0, (ptr<snapshot>)0);
    }
  }
}

void
snapshot::fmap_entry_load_cb (chordID sid, int err)
{
  if (err) {
    warn << "fmap_entry_load_cb: cannot insert into db, error "
         << strerror(err) << " (" << err << ")\n";
    fmap_entry_insert_failures++;
  }
  assert(fmap_entry_inserts > 0);
  fmap_entry_inserts--;
  if (fmap_entry_inserts == 0 && fmap_load_done) {
    if (fmap_entry_insert_failures == 0)
      bcb (sid, mkref(this));
    else
      bcb (0, (ptr<snapshot>)0);
  }
}

snapshot::snapshot
(chordID vid, chordID prev, ptr<snapshot> image, cb_build cb)
  : vid(vid), bcb(cb)
{
  insert_failures = 0;
  outstanding_inserts = 0;
  outstanding_treeops = 0;
  inserts = 0;
  fmap_db = 0;
  id = 0;
  tree_build_started = false;
    
  if (!(fmap_db = fmap_createdb ())) {
    warn << "cannot create fmap database.\n";
    timecb (tsnow, wrap (mkref(this), &snapshot::snapshot_failed));
    return;
  }
  fmap_entry_inserts = 0;
  fmap_entry_insert_failures = 0;
  fmap_load_done = false;

  if (image) {
    prev_link = prev_snapshot = prev;

    ptr<dbEnumeration> it = image->fmap_db->enumerate();
    ptr<dbPair> p;
    while (p = it->nextElement()) {
      chordID inum = dbrec2id(p->key);
      chordID chash = dbrec2id(p->data);
      ptr<dbrec> k = id2dbrec (inum);
      ptr<dbrec> d = id2dbrec (chash);
      fmap_entry_inserts++;
      fmap_db->insert
	(k, d,
	 wrap (mkref(this), &snapshot::fmap_entry_insert_from_prev_cb, image));
    }
    fmap_load_done = true;
  }
  else {
    prev_link = prev_snapshot = 0;
    build_from_scratch();
  }
}

void
snapshot::fmap_entry_insert_from_prev_cb (ptr<snapshot> image, int err)
{
  if (err) {
    warn << "fmap_entry_load_cb: cannot insert into db, error "
         << strerror(err) << " (" << err << ")\n";
    fmap_entry_insert_failures++;
  }
  assert(fmap_entry_inserts > 0);
  fmap_entry_inserts--;
  if (fmap_entry_inserts == 0 && fmap_load_done) {
    if (fmap_entry_insert_failures == 0)
      build_from_prev (image);
    else
      bcb (0, (ptr<snapshot>)0);
  }
}

void
snapshot::snapshot_failed ()
{
  bcb (0, (ptr<snapshot>)0);
}

ptr<dbfe>
snapshot::fmap_createdb()
{
  assert(fmap_db == 0);

  ptr<dbfe> db = New refcounted<dbfe>();
  db->set_compare_fcn (wrap (&dbcompare));
  dbOptions opts;
  opts.addOption("opt_async", 1);
  opts.addOption("opt_flag", DB_CREATE | DB_TRUNCATE);

  // don't supply a path: sleepycat DB3 is not async. to avoid
  // deadlock we need to keep the DB in-memory. furthermore, don't
  // want to corrupt the fmap_db of other snapshots.

  if (int err = db->opendb(0L, opts)) {
    warn << "open returned: " << strerror(err) << " (" << err << ")\n";
    return 0;
  }

  return db;
}

void 
snapshot::del ()
{
  active_snodes.deleteall ();
  tole.clear ();
  if (fmap_db)
    fmap_db = 0;
  _version.clear ();
}

void 
snapshot::build_from_scratch()
{
  myview = New refcounted<view> (wrap(mkref(this), &snapshot::got_view), vid);
}

void
snapshot::got_view(cfsrw_status status)
{
  if (status) {
    warn << "can't load view\n"; 
    bcb (0, (ptr<snapshot>)0);
    return;
  }
  rootdir = myview->root ();
  vervec svec;
  vNew refcounted<logiter>
    (wrap(mkref(this), &snapshot::update_tole, svec), vid);
}

void
snapshot::build_from_prev(ptr<snapshot> ps)
{
  rootdir = ps->rootdir;
  assert (vid == ps->view_addr ());
  vNew refcounted<logiter>
    (wrap (mkref(this), &snapshot::update_tole, ps->version ()), vid);
}

void 
snapshot::update_tole (vervec svec, iterinfo *li)
{
  bool stop = false;
  if (!li || li->status != CFSRW_OK) {
    warn << "snapshot::update_tole: error!\n";
    bcb (0, (ptr<snapshot>)0);
    return;
  }
  if (!li->finished) {
    if (prev_snapshot > 0) {
      for (uint j=0; j<svec.size(); j++) {
	if (li->record->head () == svec.head(j)) {
	  if (li->record->seqnum () == svec.seqnum(j)) {
	    li->eol = true;
	    stop = true;
	  }
	  break;
	}
      }
    }
    if (!stop) {
      WARN(2) << "tole: head " << id2str(li->record->head ()) << "\n";
      WARN(2) << "tole: this " << id2str(li->record->hash ()) << "\n";
      WARN(2) << "tole: seqn " << li->record->seqnum () << "\n";
      WARN(2) << "tole: type " << li->record->record_type () << "\n";
      tole.push_back (li->record);
    }
  }
  
  else if (!tree_build_started) {
    tree_build_started = true;
    warn << "snapshot: " << tole.size() << " records to process\n";
    _version = li->iter.version ();
    outstanding_treeops = 0;
    inode_stored = 0;
    treeop_failures = 0;
    tree_stored = false;
    fs_tree();
  }
}

void 
snapshot::fs_tree()
{
  assert(outstanding_treeops == 0);
  if (outstanding_treeops == 0)
    timecb (tsnow, wrap(mkref(this), &snapshot::build_tree));
}

void
snapshot::build_tree()
{
  // build a filesystem tree one operation at a time: avoid concurrent
  // building so we don't have to use locks! this may be slow...

  if (!tole.empty() && !treeop_failures) {
    ptr<log_record> e = tole.pop_back ();
    WARN(2) << "processing log type: " << e->record_type () << "\n";
    switch (e->record_type ()) {
    case log_record::T_INODE:
      {
        inode_log_record *ie = e->template record<inode_log_record>();
        ptr<sfsro_data> sdata = New refcounted<sfsro_data>(SFSRO_INODE);
        setinode (e, sdata->inode);
        update_index (ie->object (), sdata);
	if (outstanding_treeops == 0)
	  fs_tree();
      }
      break;
    case log_record::T_HEAD:
      {
        head_log_record *hle = e->template record<head_log_record>();
        rootdir = hle->root ();
	fs_tree();
      }
      break;
    case log_record::T_LINK:
      {
        link_log_record *le = e->template record<link_log_record>();
        do_link (le->dir (), le->name (), e);
      }
      break;
    case log_record::T_UNLINK:
      {
        unlink_log_record *ue = e->template record<unlink_log_record>();
        do_unlink (0, ue->dir (), ue->name (), e);
      }      
      break;
    case log_record::T_WRITE:
      do_write (e);
      break;
    case log_record::T_ATTR:
      update_attr (e);
      break;
    case log_record::T_RENAME:
      do_rename (e);
      break;

    case log_record::T_END:
    case log_record::T_INTEND:
    default:
      fs_tree();
      break;
    }
  }
  else if (!treeop_failures && outstanding_treeops == 0 && 
           tole.empty() && !tree_stored) {
    tree_stored = true;
    WARN(2) << "store tree\n";
    store_tree();
  }
  else if (treeop_failures) {
    warn << "snapshot built failed! dropping this snapshot.\n";
    // for debugging
    // fatal << "snapshot failed.\n";
    bcb (0, (ptr<snapshot>)0);
  }
}

void
snapshot::store_tree()
{
  if (outstanding_treeops == 0) {
    if (prev_snapshot == 0) {
      snode *se = active_snodes[rootdir];
      assert (se);
      se->ent->inode->reg->nlink++; // rootdir nlink is always 1
      store (rootdir);
    }
    else if (active_snodes.size() > 0) {
      active_snodes.traverse (wrap (mkref(this), &snapshot::store_new));
      if (outstanding_inserts == 0)
        insert_snapshot ();
    }
    else
      insert_snapshot();
  }
}

void 
snapshot::setinode (ptr<log_record> e, sfsro_inode *inode)
{
  inode_log_record *ie = e->template record<inode_log_record>();
  strbuf w;
  sid (w, ie->object ());
  WARN(2) << "setinode: inode type: " << ie->object_type ()
          << " inum: " << armor64A(w) << "\n";

  switch (ie->object_type ()) {
  case CFSRW_DIR:    // ignore dir opaque case
  case CFSRW_REG:
    {
      ftypero t =
	(ie->object_type () == CFSRW_DIR) ? SFSRODIR : SFSROREG;
      inode->set_type (t); 
      rpc_clear (*inode->reg);
      inode->reg->path = armor64A(w);
      inode->reg->nlink = 0;
      
      init_attr (inode->reg);
      update_ctime (inode->reg, ie->timestamp ());
      update_mtime (inode->reg, ie->timestamp ());
      set_attr (inode->reg, ie->attr ());

      inode->reg->direct.setsize (0);
      if (t == SFSROREG)
        write_file (e, inode->reg);
    }
    break;
  case CFSRW_LNK:
    {
      inode->set_type (SFSROLNK);
      rpc_clear (*inode->lnk);
      inode->lnk->path = armor64A(w);
      inode->lnk->nlink = 0;
  
      init_attr (inode->lnk);
      update_ctime (inode->lnk, ie->timestamp ());
      set_attr (inode->lnk, ie->attr ());
      inode->lnk->dest = ""; // will be set to a path later by the write logent
    }
    break;
  default:
    warn << "snapshot::setinode: strange inode type.\n";
  }
}

void
snapshot::lookup_fmap (chordID inum, cb_snode cb)
{
  ptr<dbrec> q = id2dbrec(inum);
  fmap_db->lookup(q, wrap(mkref(this), &snapshot::lookup_fmap_cb, inum, cb));
}

void
snapshot::lookup_fmap_cb (chordID inum, cb_snode cb, ptr<dbrec> v)
  // look for the inum->chash mapping, then load the block chash
  // points to. if inum->chash mapping is missing, treat inum as a
  // content hash
{
  chordID chash = inum;
  if (v)
    chash = dbrec2id (v);
  cfsrw_dhashclt::retrieve_sfsro
    (chash, wrap (mkref(this), &snapshot::lookup_fmap_load_block_cb, cb, inum));
}

void
snapshot::lookup_fmap_load_block_cb 
(cb_snode cb, chordID inum, ptr<sfsro_data> sdata)
{
  if (sdata) {
    snode *e = update_index (inum, sdata);
    cb (e);
  }
  else
    cb ((snode*)0);
}

void 
snapshot::do_link (chordID dir, str name, ptr<log_record> eptr)
{
  outstanding_treeops++;
  WARN(2) << "do_link: dir inum: " << id2str (dir)
          << " file name: " << name << "\n";
  snode *d = active_snodes[dir];
  if (!d) {
    lookup_fmap
      (dir, wrap(mkref(this), &snapshot::do_link_got_dir, name, eptr));
  }
  else
    do_link_got_dir(name, eptr, d);
}

void
snapshot::do_link_got_dir (str name, ptr<log_record> eptr, snode *d)
{
  if (!d || !d->ent || d->ent->type != SFSRO_INODE ||
      d->ent->inode->type != SFSRODIR) {
    warn << "snapshot::do_link, no dir snode\n";
    treeop_failures++;
    outstanding_treeops--;
    fs_tree();
    return;
  }

  assert (eptr->dirop_record());
  snode *e = active_snodes[eptr->dirop_record()->object()];

  if (!e) {
    lookup_fmap
      (eptr->dirop_record()->object(),
       wrap(mkref(this), &snapshot::do_link_got_node, name, eptr, d));
  }
  else
    do_link_got_node(name, eptr, d, e);
}

void
snapshot::do_link_got_node
(str name, ptr<log_record> eptr, snode *d, snode *e)
{
  if (!e || !e->ent || e->ent->type != SFSRO_INODE) {
    warn << "snapshot::do_link: node does not exist.\n";
    treeop_failures++;
    outstanding_treeops--;
    fs_tree();
    return;
  }

  if (e->ent->inode->type == SFSRODIR) {
    // linking in a directory, set that directory's ".." reference
    strbuf s;
    sid (s, d->inum);
    memmove (e->ent->inode->reg->parent.base(), str(s).cstr(), SFSRO_FHSIZE);
  }

  sfsro_inode_reg *dnode = d->ent->inode->reg;    
  chordID dirblk;
  strbuf ds; 
  
  if (dnode->direct.size () == 0) {
    // create a new block for this directory
    ptr<sfsro_data> dir = New refcounted<sfsro_data>(SFSRO_DIRBLK);
    get_random_ID (&dirblk, SFSRO_FHSIZE);
    sid (ds, dirblk);
    update_index (dirblk, dir);
    dnode->direct.setsize (1);
    memmove (dnode->direct[0].base (), str(ds).cstr (), SFSRO_FHSIZE);
    dir->dir->eof = true;
    dir->dir->count = 0;
    add_dirent (name, eptr, dnode, dir, e);
    outstanding_treeops--;
    fs_tree();
    return;
  }

  else {
    int blk = dnode->direct.size ()-1;
    mpz_set_rawmag_be (&dirblk, dnode->direct[blk].base (), SFSRO_FHSIZE);
    sid (ds, dirblk);

    WARN(1) << "do_link: searching dnode->direct[" << blk << "] = "
            << dirblk << "\n";

    snode *dbe = active_snodes[dirblk];
    if (dbe)
      do_link_got_dirblk(name, eptr, dirblk, d, e, dbe);
    else {
      lookup_fmap
	(dirblk,
	 wrap(mkref(this), &snapshot::do_link_got_dirblk, name, eptr,
	      dirblk, d, e));
    }
  }
}

void
snapshot::do_link_got_dirblk
(str name, ptr<log_record> eptr, chordID dirblk,
 snode *d, snode *e, snode *dbe)
{
  if (!dbe || dbe->ent->type != SFSRO_DIRBLK) {
    warn << "snapshot::do_link: dirblk does not exist.\n";
    treeop_failures++;
    outstanding_treeops--;
    fs_tree();
    return;
  }
  
  ptr<sfsro_data> dir = dbe->ent;
  sfsro_inode_reg *dnode = d->ent->inode->reg;

  add_dirent (name, eptr, dnode, dir, e);
  outstanding_treeops--;
  fs_tree();
}

void
snapshot::add_dirent (str name, ptr<log_record> e, sfsro_inode_reg *dnode,
                      ptr<sfsro_data> dir, snode *fe)
{
  // directory size is bytes needed to store directory, so we set that
  // before we insert directory block into dhash, not here.

  if (dir->dir->count == SNAPSHOT_DBLKSIZE) {
    warn << "snapshot directory size limit reached!!!\n";
    treeop_failures++;
    return;
  }

  chordID inum;
  assert (e->dirop_record());
  update_ctime (dnode, e->dirop_record()->timestamp ());
  inum = e->dirop_record()->object();

  rpc_ptr<sfsro_dirent> direntp;
  direntp.alloc ();
  strbuf ds;
  sid(ds, inum);

  memmove (direntp->fh.base (), str(ds).cstr (), SFSRO_FHSIZE);
  WARN(2) << "adding to dirent " << name << ", inum " 
          << id2str (inum) << "\n";
  direntp->fileid = cfsfh2fid (inum);
  direntp->name = name;

  if (dir->dir->entries)
    direntp->nextentry = dir->dir->entries;
  dir->dir->entries = direntp;
  dir->dir->count++;
    
  if (dir->dir->count == SNAPSHOT_DBLKSIZE) {
    int blk = dnode->direct.size ();
    if (blk < SFSRO_NDIR) {
      dir->dir->eof = false;
      // create new directory block
      chordID dirblk;
      ptr<sfsro_data> newdir = New refcounted<sfsro_data>(SFSRO_DIRBLK);
      get_random_ID (&dirblk, SFSRO_FHSIZE);
      update_index (dirblk, newdir);
      newdir->dir->eof = true;
      newdir->dir->count = 0;
      WARN(1) << "created new dir blk (" << blk << ") " << dirblk << "\n";

      rpc_vec<sfs_hash, SFSRO_NDIR> dv = dnode->direct;
      dnode->direct.setsize (blk+1);
      for (int j=0; j<blk; j++) dnode->direct[j] = dv[j];
      strbuf ds;
      sid (ds, dirblk);
      memmove (dnode->direct[blk].base (), str(ds).cstr (), SFSRO_FHSIZE);
    }
  }

  sfsro_inode *inode = (fe->ent)->inode;
  if (inode->type == SFSROLNK)
    (*(*inode).lnk).nlink++;
  else {
    (*(*inode).reg).nlink++;
    if (inode->type == SFSRODIR) // take into account of ..
      dnode->nlink++;
  }
}

snode * 
snapshot::update_index (chordID inum, ptr<sfsro_data> sdata)
{
  snode *se = New snode (inum, sdata);
  active_snodes.insert (se);
  WARN(2) << "snapshot::update_index: inum " << id2str (inum) << "\n";
  return se;
}

void 
snapshot::do_write(ptr<log_record> e)
{
  outstanding_treeops++;
  write_log_record *we = e->template record<write_log_record>();
  snode *f = active_snodes[we->file()];
  if (f)
    do_write_got_node(e, f);
  else
    lookup_fmap
      (we->file(), wrap(mkref(this), &snapshot::do_write_got_node, e));
}

void
snapshot::do_write_got_node(ptr<log_record> e, snode *f)
{
  if (!f || f->ent->type != SFSRO_INODE) {
    warn << "snapshot::do_write: file inode does not exist.\n";
    treeop_failures++;
    outstanding_treeops--;
    fs_tree();
    return;
  }

  sfsro_inode *inode = (f->ent)->inode;
  if (inode->type == SFSROLNK) {
    do_symlink (e, inode->lnk);
    outstanding_treeops--;
    fs_tree();
  }
  else if (inode->type == SFSROREG || inode->type == SFSROREG_EXEC) {
    outstanding_treeops--;
    write_file (e, inode->reg);
  }
  else {
    warn << "do_write_got_node: wrong file type!\n";
    treeop_failures++;
    outstanding_treeops--;
    fs_tree();
  }
}

void 
snapshot::do_symlink(ptr<log_record> e, sfsro_inode_lnk *lnode)
{
  write_log_record *we = e->template record<write_log_record>();
  if (we->count () > MAXPATHLEN) {
    warn << "snapshot::do_symlink: target too large.\n";
    treeop_failures++;
    return;
  }
  lnode->dest = nfspath3 (we->data ());
  update_ctime (lnode, we->timestamp ());
}

void 
snapshot::write_file (ptr<log_record> e, sfsro_inode_reg *rnode)
{
  outstanding_treeops++;
  str fname = get_filename (rnode->path);
  snapshot::a->open (fname, O_WRONLY | O_CREAT, 0666,
                     wrap (mkref(this), &snapshot::write_file_open_cb,
		           e, rnode));
}

void
snapshot::write_file_open_cb (ptr<log_record> e, sfsro_inode_reg *rnode,
                              ptr<aiofh> fh, int err)
{
  if (!fh) {
    warn << "write_file: open failed: " << err << "\n";
    treeop_failures++;
    outstanding_treeops--;
    fs_tree();
    return;
  }

  write_log_record *we = e->template record<write_log_record> ();
  inode_log_record *ie = e->template record<inode_log_record> ();
  attr_log_record  *ae = e->template record<attr_log_record> ();

  if (we) {
    ptr<aiobuf> buf = snapshot::a->bufalloc (we->count ());
    if (!buf) {
      warn << "no buf, bufwait\n";
      snapshot::a->bufwait
	(wrap (mkref(this), &snapshot::write_file_open_cb, e, rnode, fh, 0));
      return;
    }

    // mark region as modified
    if (rnode->mstart == 0 && rnode->mend == 0) {
      rnode->mstart = we->offset ();
      rnode->mend = we->offset () + we->count ();
      WARN(2) << "mstart set to " << rnode->mstart
	      << " and mend set to " << rnode->mend << "\n";
    }
    else {
      if (rnode->mstart > we->offset ())
	rnode->mstart = we->offset ();
      if (rnode->mend < we->offset () + we->count ())
	rnode->mend = we->offset () + we->count ();
      WARN(2) << "mstart set to " << rnode->mstart
	      << " and mend set to " << rnode->mend << "\n";
    }
    // write beyond size of file, mark everything between old size of
    // file to where write started as modified
    if (rnode->attr.size < rnode->mstart)
      rnode->mstart = rnode->attr.size;

    memmove(buf->base (), we->data (), we->count ());
    WARN(2) << "calling aiod::write at offset " << we->offset() << "\n";
    fh->write (we->offset (), buf,
	       wrap (mkref(this), &snapshot::write_file_write_cb,
		     e, rnode, fh));
    return;
  }
  else if (ie || ae) {
    if ((ie && ie->attr ()->valid_size) || (ae && ae->attr ()->valid_size)) {
      uint64 size = ie ? ie->attr ()->size : ae->attr ()->size;

      // mark the whole file as modified in case the setattr increased
      // size of the file
      rnode->mstart = 0;
      rnode->mend = size;
      WARN(2) << "mstart set to " << rnode->mstart
	      << " and mend set to " << rnode->mend << "\n";

      WARN(2) << "aiod::truncate file to " << size << " bytes\n";
      fh->ftrunc (size, wrap (mkref(this), &snapshot::write_file_trunc_cb,
	                      e, rnode, fh));
    }
    else
      write_file_write_cb (e, rnode, fh, 0, 0, 0); // close file
    return;
  }
  assert (0);
}

void
snapshot::write_file_trunc_cb (ptr<log_record> e, sfsro_inode_reg *rnode,
                               ptr<aiofh> fh, int err)
{
  if (err) {
    warn << "write_file: trunc failed: " << err << "\n";
    treeop_failures++;
  }
  fh->close (wrap (mkref(this), &snapshot::write_file_close_cb));
}

void
snapshot::write_file_write_cb (ptr<log_record> e, sfsro_inode_reg *rnode,
                               ptr<aiofh> fh, ptr<aiobuf> buf,
			       ssize_t sz, int err)
{
  write_log_record *we = e->template record<write_log_record>();
  if (we) {
    if (err || ((unsigned) sz != we->count ())) {
      warn << "write_file: write failed: " << err << "\n";
      treeop_failures++;
    }
    update_mtime (rnode, we->timestamp ());
    update_ctime (rnode, we->timestamp ());

    if (rnode->attr.size < (we->offset () + we->count ()))
      rnode->attr.size = we->offset () + we->count ();
  }
  fh->close (wrap (mkref(this), &snapshot::write_file_close_cb));
}

void
snapshot::write_file_close_cb (int err)
{
  if (err) {
    warn << "write_file: close failed: " << err << "\n";
    treeop_failures++;
  }
  outstanding_treeops--;
  fs_tree();
}

void 
snapshot::do_unlink(int phase, chordID dir, str name, ptr<log_record> e)
{
  outstanding_treeops++;
  snode *d = active_snodes[dir];
  if (d)
    do_unlink_got_dir(phase, name, e, d);
  else {
    lookup_fmap
      (dir, wrap(mkref(this), &snapshot::do_unlink_got_dir, phase, name, e));
  }
}

void
snapshot::do_unlink_got_dir(int phase, str name, ptr<log_record> e, snode* d)
{
  if (!d || d->ent->inode->type != SFSRODIR) {
    warn << "snapshot::do_unlink: no SFSRODIR entry.\n";
    treeop_failures++;
    outstanding_treeops--;
    fs_tree();
    return;
  }
  sfsro_inode_reg *dnode = d->ent->inode->reg;
  if (dnode->direct.size () == 0) {
    done_unlink(phase, e, true);
    return;
  } 
  
  chordID dirblk;
  mpz_set_rawmag_be (&dirblk, dnode->direct[0].base (), SFSRO_FHSIZE);
  snode *dbe = active_snodes[dirblk];
  if (dbe)
    do_unlink_got_dirblk(phase, name, e, 0, d, dbe);
  else
    lookup_fmap
      (dirblk, wrap(mkref(this), &snapshot::do_unlink_got_dirblk,
		    phase, name, e, 0, d));
}

void
snapshot::do_unlink_got_dirblk (int phase, str name, ptr<log_record> e,
                                int blknum, snode *d, snode *dbe)
{
  if (!dbe || dbe->ent->type != SFSRO_DIRBLK) {
    warn << "snapshot::do_unlink: dirblk does not exist.\n";
    treeop_failures++;
    outstanding_treeops--;
    fs_tree();
    return;
  }
  sfsro_data *dir = dbe->ent;
  remove_dirent (phase, name, e, d, dir, blknum);
}

void 
snapshot::remove_dirent (int phase, str name, ptr<log_record> e,
                         snode *d, sfsro_data *dir, uint32 blknum)
{
  sfsro_inode_reg *dnode = d->ent->inode->reg;
  assert(e->dirop_record());
  update_ctime (dnode, e->timestamp ());

  chordID fhID;
  sfsro_dirent* de = dir->dir->entries;
  if (de && (de->name == name)) {
    mpz_set_rawmag_be (&fhID, de->fh.base (), SFSRO_FHSIZE);
    dir->dir->entries = dir->dir->entries->nextentry;
  }
  else {
    while (de && de->nextentry && (de->nextentry->name != name))
      de = de->nextentry;
    if (!de || !de->nextentry) {
      if (blknum == dnode->direct.size ()-1)
        done_unlink(phase, e, true);
      else {
        blknum++;
	chordID dirblk;
        mpz_set_rawmag_be
	  (&dirblk, dnode->direct[blknum].base (), SFSRO_FHSIZE);
        snode *dbe = active_snodes[dirblk];
        if (dbe)
          do_unlink_got_dirblk(phase, name, e, blknum, d, dbe);
        else
          lookup_fmap
            (dirblk, wrap(mkref(this), &snapshot::do_unlink_got_dirblk,
		          phase, name, e, blknum, d));
      }
      return;
    }
    mpz_set_rawmag_be (&fhID, de->nextentry->fh.base (), SFSRO_FHSIZE);
    de->nextentry = de->nextentry->nextentry;
  }
  assert(dir->dir->count > 0);
  dir->dir->count--;

  snode *ie = active_snodes[fhID];
  if (ie)
    remove_dirent_got_fh(phase, fhID, e, dnode, ie);
  else {
    lookup_fmap
      (fhID, wrap(mkref(this), &snapshot::remove_dirent_got_fh,
	          phase, fhID, e, dnode));
  }
}

void
snapshot::remove_dirent_got_fh
(int phase, chordID fhID, ptr<log_record> e, sfsro_inode_reg *dnode,
 snode *ie)
{
  if (!ie || ie->ent->type != SFSRO_INODE) {
    warn << "snapshot::remove_dirent: entry is not type INODE.\n";
    treeop_failures++;
  }
  else {
    sfsro_inode *inode = (ie->ent)->inode;
    if (inode->type == SFSROLNK)
      (*(*inode).lnk).nlink--;
    else {
      (*(*inode).reg).nlink--;
      if (inode->type == SFSRODIR) // take into account of ..
        dnode->nlink--;
    }
  }
  done_unlink(phase, e);
}

void
snapshot::done_unlink(int phase, ptr<log_record> e, bool error)
{
  outstanding_treeops--;
  rename_log_record *re = e->template record<rename_log_record>();  
  if (re) {
    if (phase == 0)
      do_unlink (1, re->from_dir(), re->from_name(), e);
    else
      do_link (re->dir(), re->name (), e);
  }
  else {
    if (error) {
      warn << "unlink failed\n";
      treeop_failures++;
    }
    fs_tree();
  }
}

void 
snapshot::update_attr (ptr<log_record> eptr)
{
  outstanding_treeops++;
  attr_log_record *ae = eptr->template record<attr_log_record>();
  snode *e = active_snodes[ae->object ()];
  if (e)
    update_attr_got_node (eptr, e);
  else {
    lookup_fmap
      (ae->object(), wrap(mkref(this), &snapshot::update_attr_got_node, eptr));
  }
}

void
snapshot::update_attr_got_node (ptr<log_record> eptr, snode *e)
{
  attr_log_record *ae = eptr->template record<attr_log_record>();
  if (!e || e->ent->type != SFSRO_INODE) {
    warn << "snapshot::update_attr: type is not INODE.\n";
    treeop_failures++;
    outstanding_treeops--;
    fs_tree ();
  }
  else {
    if (e->ent->inode->type == SFSROLNK) {
      update_ctime (e->ent->inode->lnk, ae->timestamp ());
      set_attr (e->ent->inode->lnk, ae->attr ());
      outstanding_treeops--;
      fs_tree ();
    }
    else {
      update_ctime (e->ent->inode->reg, ae->timestamp ());
      if(ae->attr()->valid_size)
        update_mtime (e->ent->inode->reg, ae->timestamp ());
      set_attr (e->ent->inode->reg, ae->attr ());
      if (e->ent->inode->type == SFSROREG) {
        outstanding_treeops--;
        write_file (eptr, e->ent->inode->reg);
      }
      else {
        outstanding_treeops--;
        fs_tree ();
      }
    }
  }
}

void 
snapshot::do_rename (ptr<log_record> e)
{
  rename_log_record *re = e->template record<rename_log_record>();  
  do_unlink (0, re->dir(), re->name(), e);
}

