/*
 *
 * Copyright (C) 2002 Athicha Muthitacharoen (athicha@lcs.mit.edu)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "debug.h"
#include "snapshot.h"

void
snapshot::fmap_del_cb (int err)
{
  if (err && err != DB_NOTFOUND) {
    WARN(1) << "removing entry from fmap: " << err << "\n";
  }
}

void 
snapshot::store_new (snode *e)
{
  if (e->ent->type == SFSRO_INODE) {
    sfsro_inode *inode = e->ent->inode;
    if ((inode->type == SFSROLNK && inode->lnk->nlink > 0) || 
	(inode->type != SFSROLNK && inode->reg->nlink > 0))
      store (e->inum);
    else {
      ptr<dbrec> q = id2dbrec(e->inum);
      fmap_db->del (q, wrap (mkref(this), &snapshot::fmap_del_cb));
    }
  }
}

void 
snapshot::store (chordID cID) 
{
  WARN(1) << "store: inum = " << id2str (cID) << "\n";
  snode *e = active_snodes[cID];
  if (!e) {
    warn << "snapshot::store: entry does not exist in active_snodes.\n";
    insert_failures++;
    return;
  }
  switch (e->ent->type) {
  case SFSRO_INODE: {
    ftypero type = e->ent->inode->type;
    sfsro_inode *inode = e->ent->inode;
    if (type == SFSROLNK)
      store_inode (cID, inode);
    else if ((type == SFSROREG || type == SFSROREG_EXEC))
      store_file (cID, inode);
    else
      store_directory (cID, inode);
  }
  break;
  default:
    warn << "entry type should not have been in active_snodes.\n";
    insert_failures++;
  }
}

void
snapshot::insert (chordID hash, char *data, size_t len, ptr<sfsro_data> d)
{
  outstanding_inserts++;
  inserts++;
  cfsrw_dhashclt::insert_sfsro
    (hash, data, len, wrap (mkref(this), &snapshot::insert_cb, false, 0), d);
  xfree (data);
}

void
snapshot::insert_and_addtofmap (char *data, size_t len, chordID inum,
                                ptr<sfsro_data> d)
{
  outstanding_inserts++;
  inserts++;
  cfsrw_dhashclt::insert_sfsro
    (data, len, wrap (mkref(this), &snapshot::insert_cb, true, inum), d);
  xfree (data);
}

void 
snapshot::insert_cb (bool map, chordID inum, dhash_stat status, chordID key)
{
  outstanding_inserts--;
  if (status == DHASH_OK) {
    if (map) {
      // insert inum->chash to fmap
      ptr<dbrec> k = id2dbrec (inum);
      ptr<dbrec> d = id2dbrec (key);
      outstanding_inserts++;
      fmap_db->insert
        (k, d, wrap(mkref(this), &snapshot::fmap_entry_insert_on_store_cb));
      return;
    }
  }
  else {
    warn << "snapshot::insert_cb: dhash insert failed.\n";
    insert_failures++;
  }
  
  if (outstanding_inserts == 0)
    insert_snapshot ();
}

void
snapshot::fmap_entry_insert_on_store_cb (int err)
{
  outstanding_inserts--;
  if (err) {
    warn << "fmap_entry_insert_on_store_cb: cannot insert into db, error "
         << strerror(err) << " (" << err << ")\n";
    insert_failures++;
  }
  if (outstanding_inserts == 0)
    insert_snapshot ();
}

void 
snapshot::insert_snapshot ()
{
  WARN(1) << "snapshot: " << inserts << " sfsro blocks inserted\n";
  if (insert_failures)
    bcb (0, (ptr<snapshot>)0);
  else
    serialize_and_store();
}

chordID
snapshot::store_one_fmap_block (str s)
{
  fmap_block_inserts++;
  chordID n = compute_hash (s.cstr (), s.len ());
  cfsrw_dhashclt::insert_ch
    (n, s.cstr (), s.len (),
     wrap (mkref(this), &snapshot::insert_fmapblock_cb));
  return n;
}

str
snapshot::serialize_info ()
{
  strbuf sb;
  sid (sb, prev_link);
  sid (sb, id);
  return sb;
}

void
snapshot::serialize_and_store ()
{
  fmap_block_inserts = 0;
  fmap_block_insert_failures = 0;

  vec<chordID> inum;
  vec<chordID> chash;
  ptr<dbEnumeration> it = fmap_db->enumerate();
  ptr<dbPair> p;
  while (p = it->nextElement()) {
    inum.push_back(dbrec2id(p->key));
    chash.push_back(dbrec2id(p->data));
  }

  chordID next = 0;
  unsigned total = 0;
  while (inum.size () > FMAP_ENTRIES_PER_PAGE) {
    unsigned n=0;
    strbuf sb;
    sint (sb, FMAP_ENTRIES_PER_PAGE);  // entries in this block
    sint (sb, total);                  // remaining entries after this block
    sid  (sb, next);                   // address of next block
    while (n < FMAP_ENTRIES_PER_PAGE) {
      chordID i = inum.pop_back ();
      chordID c = chash.pop_back ();
      sid (sb, i);
      sid (sb, c);
      total++;
      n++;
    }
    next = store_one_fmap_block (sb);
    WARN(1) << "stored extra fmap page at " << next << "\n";
  }

  strbuf sb;
  sid (sb, prev ());
  sid (sb, view_addr ());
  sid (sb, rootdir);
  _version.serialize(sb);
  sint (sb, inum.size ()); // entries in this block
  sint (sb, total);        // remaining entries after this block
  sid  (sb, next);         // address of next block
  while (inum.size () > 0) {
    chordID i = inum.pop_back ();
    chordID c = chash.pop_back ();
    sid (sb, i);
    sid (sb, c);
  }
  id = store_one_fmap_block (sb);
}

void 
snapshot::insert_fmapblock_cb (dhash_stat status, chordID key)
{
  assert (fmap_block_inserts > 0);
  fmap_block_inserts--;
  if (status != DHASH_OK)
    fmap_block_insert_failures++;

  if (fmap_block_inserts == 0) {
    if (!fmap_block_insert_failures) {
      WARN(2) << "stored snapshot: " << id << ", now try snapinfo\n";
      str snapinfo = serialize_info ();
      cfsrw_dhashclt::insert_ch
        (snapinfo.cstr (), snapinfo.len (),
         wrap (mkref(this), &snapshot::insert_snapinfo_cb));
    }
    else {
      warn << "snapshot::insert_fmapblock_cb: dhash insert failed.\n";
      bcb (0, (ptr<snapshot>)0);
    }
  }
}

void 
snapshot::insert_snapinfo_cb (dhash_stat status, chordID key)
{
  if (status == DHASH_OK) {
    WARN(2) << "stored snapinfo!! snapinfo id: " << key << "\n";
    WARN(1) << "snapshot stored, did " << inserts << " dhash insertions\n";
    bcb (key, mkref(this));
  } else {
    warn << "snapshot::insert_snapinfo_cb: dhash insert failed.\n"; 
    bcb (0, (ptr<snapshot>)0);
  }
}

void
snapshot::store_inode (chordID inum, sfsro_inode *inode) 
{
  inode_stored++;

  ptr<sfsro_data> dat = New refcounted<sfsro_data> (SFSRO_INODE);
  *(dat->inode) = *inode;

  size_t calllen = 0;
  char *callbuf = NULL;
  xdrsuio x (XDR_ENCODE);
  if (xdr_sfsro_data (x.xdrp (), dat)) {
    calllen = x.uio ()->resid ();
    callbuf = suio_flatten (x.uio ());
  }
  insert_and_addtofmap (callbuf, calllen, inum, dat);
}

void 
snapshot::store_directory (chordID inum, sfsro_inode *inode)
{
  WARN(1) << "storing directory " << id2str(inum) << "\n";
  unsigned blk = 0;
  unsigned dirsize = 0;

  for (blk = 0; blk < inode->reg->direct.size(); blk++) {
    chordID dID;
    mpz_set_rawmag_be (&dID, inode->reg->direct[blk].base (), SFSRO_FHSIZE);  
    snode *dirblk_ent = active_snodes[dID];
    if (!dirblk_ent) // dirblk not modified or loaded
      continue;

    if (dirblk_ent->ent->type != SFSRO_DIRBLK) {
      warn << "snapshot::store_directory: type is not dirblk.\n";
      insert_failures++;
      return;
    }
    sfsro_directory *dir = dirblk_ent->ent->dir;
    sfsro_dirent *dirent = dir->entries;
    
    while (dirent) {
      mpz_set_rawmag_be (&dID, dirent->fh.base (), SFSRO_FHSIZE);
      // in build from prev, already storing everything in
      // active_snodes
      if (!prev_snapshot) 
        store (dID); 
      dirent = dirent->nextentry;
    }
  
    size_t calllen = 0;
    char *callbuf = NULL;
    xdrsuio x (XDR_ENCODE);
  
    if (xdr_sfsro_data (x.xdrp (), dirblk_ent->ent)) {
      calllen = x.uio ()->resid ();
      callbuf = suio_flatten (x.uio ());
    }
    sfs_hash chash;
    chordID n = content_hash (&chash, callbuf, calllen);
    insert (n, callbuf, calllen, dirblk_ent->ent);
    dirsize += calllen;
    inode->reg->direct[blk] = chash;
    WARN(1) << "storing block " << blk << " of directory: " << n << "\n";
  }
   
  // XXX not sure if this matters, but dirsize should be total size of
  // all dirblks, not just the total size of dirblks loaded
  inode->reg->attr.size = dirsize;
  store_inode (inum, inode);
}

void 
snapshot::store_file (chordID inum, sfsro_inode *inode)
{
  inode->reg->direct.setsize (SFSRO_NDIR);

  if (inode->reg->attr.size > 0) {
    // make sure snapshot building will not finish until we do
    outstanding_inserts++;
    str fname = get_filename (inode->reg->path);
    snapshot::a->open (fname, O_RDONLY, 0666,
	               wrap (mkref(this), &snapshot::store_file_open_cb,
			     inum, inode));
  }
  else
    store_inode (inum, inode);
}

void
snapshot::store_file_open_cb (chordID inum, sfsro_inode *inode,
                              ptr<aiofh> fd, int err)
{
  if (err) {
    warn << "cannot open snapshot file " << err << "\n";
    insert_failures++;
    outstanding_inserts--; // incremented on file open
    if (outstanding_inserts == 0)
      insert_snapshot ();
    return;
  }

  ptr<aiobuf> buf = snapshot::a->bufalloc (SNAPSHOT_BLOCKSIZE);
  if (!buf) {
    warn << "no buf, bufwait\n";
    snapshot::a->bufwait
      (wrap (mkref(this), &snapshot::store_file_open_cb,
	     inum, inode, fd, 0));
    return;
  }
  
  vec<sfs_hash>* fhs = New vec<sfs_hash>;
  fd->read (0, buf, wrap (mkref(this), &snapshot::store_file_read_block,
	                  inum, inode, 0, fhs, fd));
}

void
snapshot::store_file_read_block (chordID inum, sfsro_inode *inode,
                                 uint32 blocknum, vec<sfs_hash>* fhs,
				 ptr<aiofh> fd, ptr<aiobuf> buf,
				 ssize_t sz, int err)
{
  if (err) {
    warn << "cannot read snapshot file " << err << "\n";
    insert_failures++;
    store_file_close (inum, inode, fhs, fd);
    return;
  }
  
  blocknum++;
  bool done = false;
  if (sz < SNAPSHOT_BLOCKSIZE ||
      blocknum*SNAPSHOT_BLOCKSIZE >= inode->reg->attr.size)
    done = true;

  assert (sz >= 0);
  sfs_hash fh;
  store_file_block (inode, blocknum-1, buf->base(), sz, &fh);
  fhs->push_back (fh);

  if (done)
    store_file_close (inum, inode, fhs, fd);
  else
    fd->read (blocknum*SNAPSHOT_BLOCKSIZE, buf,
	      wrap (mkref(this), &snapshot::store_file_read_block,
		    inum, inode, blocknum, fhs, fd));
}

void 
snapshot::store_file_block (sfsro_inode *inode, uint32 blocknum,
                            const char *block, ssize_t size, sfs_hash *fh)
{
  ptr<sfsro_data> res = New refcounted<sfsro_data> (SFSRO_FILEBLK);
  res->data->setsize (size);
  memcpy (res->data->base (), block, size);

  size_t calllen = 0;
  char *callbuf = NULL;
  xdrsuio x (XDR_ENCODE);

  if (xdr_sfsro_data (x.xdrp (), res)) {
    calllen = x.uio ()->resid ();
    callbuf = suio_flatten (x.uio ());
  }

  chordID n = content_hash (fh, callbuf, calllen);
 
  if ((inode->reg->mstart == 0 && inode->reg->mend == 0) ||
      ((blocknum+1) * SNAPSHOT_BLOCKSIZE <= inode->reg->mstart) ||
      (blocknum * SNAPSHOT_BLOCKSIZE >= inode->reg->mend)) {
    WARN(2) << "block " << blocknum << " not inserted\n";
    WARN(2) << "content hash is " << n << "\n";
    xfree (callbuf);
  }
  else
    insert (n, callbuf, calllen, res);
}

void
snapshot::store_file_close_cb (int)
{
}

void
snapshot::store_file_close (chordID inum, sfsro_inode *inode,
                            vec<sfs_hash>* fhs, ptr<aiofh> fd)
{
  fd->close (wrap (mkref(this), &snapshot::store_file_close_cb));

  inode->reg->mstart = 0;
  inode->reg->mend = 0;

  unsigned blocknum = 0;
  while (blocknum < SFSRO_NDIR && blocknum < fhs->size()) {
    memmove (inode->reg->direct[blocknum].base(),
	     (*fhs)[blocknum].base(), SFSRO_FHSIZE);
    blocknum++;
  }

  if (blocknum >= SFSRO_NDIR && blocknum < SFSRO_NDIR+NFH)
    store_sind (fhs, &inode->reg->indirect, blocknum-SFSRO_NDIR, blocknum);
  if (blocknum >= SFSRO_NDIR+NFH && blocknum < SFSRO_NDIR+NFH+NFH*NFH)
    store_dind (fhs, &inode->reg->double_indirect,
	        blocknum-SFSRO_NDIR-NFH, blocknum);
  if (blocknum >= SFSRO_NDIR+NFH+NFH*NFH)
    store_tind (fhs, &inode->reg->triple_indirect,
	        blocknum-SFSRO_NDIR-NFH-NFH*NFH, blocknum);
  store_inode (inum, inode);

  delete fhs;
  outstanding_inserts --; // incremented on file open
      
  if (outstanding_inserts == 0)
    insert_snapshot ();
}

void
snapshot::store_sind (vec<sfs_hash> *fhs, sfs_hash *si_fh,
                      uint32 off, uint32& blocknum)
{
  uint32 blk = off / NFH;
  ptr<sfsro_data> sindir = New refcounted<sfsro_data> (SFSRO_INDIR);
  sindir->indir->handles.setsize (NFH);
  
  while (blk < NFH && blocknum < fhs->size()) {
    memmove (sindir->indir->handles[blk].base(),
	     (*fhs)[blocknum].base(), SFSRO_FHSIZE);
    blk++;
    blocknum++;
  }

  if (blk > 0) {
    size_t calllen = 0;
    char *callbuf = NULL;
    xdrsuio x (XDR_ENCODE);
    if (xdr_sfsro_data (x.xdrp(), sindir)) {
      calllen = x.uio ()->resid ();
      callbuf = suio_flatten (x.uio ());
    }
    chordID n = content_hash (si_fh, callbuf, calllen);
    insert (n, callbuf, calllen, sindir);
  }
}

void
snapshot::store_dind (vec<sfs_hash> *fhs, sfs_hash *di_fh,
                      uint32 off, uint32& blocknum)
{
  ptr<sfsro_data> dindir = New refcounted<sfsro_data> (SFSRO_INDIR);
  dindir->indir->handles.setsize (NFH);
  uint32 sblk = off/NFH;
  uint32 oldbn = blocknum;

  while (sblk < NFH && blocknum < fhs->size()) {
    off = (blocknum-SFSRO_NDIR)%NFH;
    store_sind
      (fhs, &dindir->indir->handles[sblk], off, blocknum);
    sblk++;
  }

  if (blocknum > oldbn) {
    size_t calllen = 0;
    char *callbuf = NULL;
    xdrsuio x (XDR_ENCODE);
    if (xdr_sfsro_data (x.xdrp(), dindir)) {
      calllen = x.uio ()->resid ();
      callbuf = suio_flatten (x.uio ());
    }
    chordID n = content_hash (di_fh, callbuf, calllen);
    insert (n, callbuf, calllen, dindir);
  }
}

void
snapshot::store_tind (vec<sfs_hash> *fhs, sfs_hash *ti_fh,
                      uint32 off, uint32& blocknum)
{
  ptr<sfsro_data> tindir = New refcounted<sfsro_data> (SFSRO_INDIR);
  tindir->indir->handles.setsize (NFH);
  uint32 dblk = off/(NFH*NFH);
  uint32 oldbn = blocknum;

  while (dblk < NFH && blocknum < fhs->size()) {
    off = blocknum%(NFH*NFH);
    store_dind
      (fhs, &tindir->indir->handles[dblk], off, blocknum);
    dblk++;
  }

  if (blocknum > oldbn) {
    size_t calllen = 0;
    char *callbuf = NULL;
    xdrsuio x (XDR_ENCODE);
    if (xdr_sfsro_data (x.xdrp(), tindir)) {
      calllen = x.uio ()->resid ();
      callbuf = suio_flatten (x.uio ());
    }
    chordID n = content_hash (ti_fh, callbuf, calllen);
    insert (n, callbuf, calllen, tindir);
  }
}

