/*
 * Test whether mtime is calculated correctly.
 * Particular issue is whether mtime reflects *last*
 * modification, rather than (for example) first.
 *
 * I believe only these operations change the mtime:
 *   create truncate write utimes
 *
 * Creates test files in current directory.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

void check_ret(int ok, char *msg);
void check_mtime(char *filename, time_t good, char *msg, char *omsg);

extern int errno;
char fname[256];

main()
{
  int fd, cc, ret, iters, what;
  time_t before, now;
  char *before_op, *op;

  srandom(getpid());

  sprintf(fname, "mtime-%d", getpid());
  fd = creat(fname, 0666);
  check_ret(fd >= 0, "creat failed");
  time(&now);
  op = "create";
  before_op = "";
  check_mtime(fname, now, op, before_op);

  for(iters = 0; iters < 20; iters++){
    sleep(2);
    check_mtime(fname, now, op, before_op);

    before = now;
    before_op = op;
    time(&now);
    op = 0;

    what = random() % 4;
    switch(what){
    case 0:
      // ret = close(fd);
      // check_ret(ret == 0, "close");
      fd = creat(fname, 0666);
      check_ret(fd >= 0, "creat");
      close(fd);
      op = "create";
      break;
    case 1:
      fd = open(fname, O_RDWR);
      check_ret(fd >= 0, "open");
      cc = write(fd, "xxx", 3);
      check_ret(cc == 3, "write");
      // check_ret(fsync(fd) == 0, "fsync");
      close(fd);
      op = "write";
      break;
    case 2:
      ret = utimes(fname, 0); /* set times to current time */
      check_ret(ret == 0, "utimes");
      op = "utimes";
      break;
    case 3:
      fd = open(fname, O_RDWR);
      ret = ftruncate(fd, random() % 5);
      check_ret(ret == 0, "ftruncate");
      close(fd);
      op = "ftruncate";
      break;
    }

    check_mtime(fname, now, op, before_op);
  }

  unlink(fname);

  exit(0);
}

void
check_ret(int ok, char *msg)
{
  if(ok == 0){
    fprintf(stderr, "mtime failure: %s: %s\n", msg, strerror(errno));
    unlink(fname);
    exit(1);
  }
}

void
check_mtime(char *file, time_t t, char *msg, char *omsg)
{
  int fd;
  struct stat sb;

  /*
   * Actually open the file to avoid client-cached attributes.
   */
  fd = open(file, 0);
  if(fd < 0){
    fprintf(stderr, "mtime: open %s: %s\n", file, strerror(errno));
    unlink(file);
    exit(1);
  }

  if(fstat(fd, &sb) != 0){
    fprintf(stderr, "mtime: fstat: %s\n", strerror(errno));
    unlink(file);
    exit(1);
  }

  if(sb.st_mtime != t &&
     sb.st_mtime != t + 1 &&
     sb.st_mtime != t - 1)
  {
    fprintf(stderr, "mtime: time check failed: %s %s\n", omsg, msg);
    fprintf(stderr, "mtime: file %s, wanted %u, got %u\n",
            file, t, sb.st_mtime);
    if(unlink(file) < 0){
      fprintf(stderr, "mtime: unlink %s failed: %s\n",
              file, strerror(errno));
    }
    exit(1);
  }

  close(fd);
}
