#!/usr/bin/perl -w

# Test interactions between multiple clients.
# You can specify two machine / username pairs, the script
# will ssh to those machines to perform individual operations.

# To run this on just one machine:
# ./twotest.pl /cfsrw/chord:xxx 127.0.0.1 root 127.0.0.1 root

# source (or whatever) /root/.cfsrw/envrun.sh
# run cfsagent on both machines
# rtm : cfslog -s ~/.sfs/identity (yields XXX)
#       utPEsfjmzBZINbbdSu+NeZ9Qbtw
# rtm2: cfslog -s ~/.sfs/identity (yields YYY)
#       jIMYA4hqXwottpu3BIyLimt825o
# rtm : viewctl -c XXX (yields ZZZ)
# rtm : viewctl -a YYY ZZZ (yields WWW)
# twotest.pl /cfsrw/WWW 127.0.0.1 rtm 127.0.0.1 rtm2

use strict;
$| = 1;

# Set to one if you want to sleep(1) before every file
# system modification. This makes up for FreeBSD NFS
# clients that check only the tv_sec part of mtime in
# their cache consistency checks.
my $sleeps = 1;

my $fs;    # "/cfsrw/chord:xxx"
my $where; # [ [ host1, user1 ], [ host2, user2 ]]

sub usage {
    print STDERR "Usage: twotest.pl /cfsrw/chord:xxx host1 user1 host2 user2\n";
    exit(1);
}

sub main {
    if($#ARGV != 4){
        usage();
    }
    $fs = $ARGV[0];
    $where = [ [ $ARGV[1], $ARGV[2] ], [ $ARGV[3], $ARGV[4] ]];

    # basic read/write
    xcreate(0, "fa$$", "zoom");
    xcheckf(1, "fa$$", "zoom");

    # open/close consistency
    xcreate(0, "fa$$", "xxx");
    xcheckf(1, "fa$$", "xxx");

    # mv
    xmv(0, "fa$$", "fb$$");
    xcheckf(1, "fb$$", "xxx");
    xchecknot(1, "fa$$");

    # truncate
    xcreate(1, "fb$$", "");
    xcheckf(0, "fb$$", "");
    xcreate(1, "fb$$", "1234567");
    xtruncate(0, "fb$$", 5);
    xcheckf(1, "fb$$", "12345");

    # unlink
    xrm(0, "fb$$");
    xchecknot(1, "fb$$");

    # mv from one directory to another
    xmkdir(1, "da$$");
    xmkdir(0, "db$$");
    xcreate(0, "da$$/xf1", "iijjkk");
    xmv(1, "da$$/xf1", "db$$");
    xcheckf(0, "db$$/xf1", "iijjkk");

    # hard link in another directory
    xlinks(1, "db$$/xf1", 1);
    xln(0, "db$$/xf1", "da$$/xx");
    xlinks(1, "db$$/xf1", 2);
    xcheckf(1, "da$$/xx", "iijjkk");
    xcreate(0, "db$$/xf1", "abcd");
    xcheckf(1, "da$$/xx", "abcd");

    # unlink one link, read through the other link
    xrm(0, "da$$/xx");
    xcheckf(1, "db$$/xf1", "abcd");
    xlinks(1, "db$$/xf1", 1);

    # clean up
    xrm(0, "db$$/xf1");
    xrmdir(0, "da$$");
    xrmdir(0, "db$$");
    xchecknot(1, "db$$/xf1");

    # destructive rename
    xcreate(0, "ff$$", "aaaa");
    xcreate(1, "fg$$", "bbbb");
    xcheckf(0, "fg$$", "bbbb");
    xmv(1, "fg$$", "ff$$");
    xcheckf(0, "ff$$", "bbbb");
    xchecknot(0, "fg$$");
    xrm(0, "ff$$");

    # mkdir is exclusive
    xmkdir(0, "dc$$");
    xcreate(0, "dc$$/f0", "f0f0");
    xmkdir(1, "dc$$"); # should fail
    xcreate(1, "dc$$/f1", "f1f1");
    xcheckf(1, "dc$$/f0", "f0f0");
    xcheckf(0, "dc$$/f1", "f1f1");
    xrm(0, "dc$$/f0");
    xrm(0, "dc$$/f1");
    xrmdir(0, "dc$$");

    # symlink
    xsymlink(0, "fc$$", "la$$");
    xchecknot(1, "la$$");
    xreadlink(1, "la$$", "fc$$");
    xcreate(0, "fc$$", "linked-file");
    xcheckf(1, "la$$", "linked-file");
    xmv(1, "la$$", "lz$$");
    xcheckf(0, "lz$$", "linked-file");
    xchecknot(0, "la$$");
    xreadlink(0, "lz$$", "fc$$");
    xrm(0, "fc$$");
    xrm(0, "lz$$");
    
    # symlink create is exclusive
    xcreate(0, "fd$$", "file-d");
    xsymlink(0, "fd$$", "lb$$");
    xcreate(1, "fe$$", "file-e");
    xsymlink(1, "fe$$", "lb$$"); # this should fail
    xcheckf(1, "lb$$", "file-d");
    xcheckf(0, "lb$$", "file-d");
    xrm(0, "fd$$");
    xrm(0, "fe$$");
    xrm(0, "lb$$");

    # output of ls
    # FreeBSD has trouble with this.
    # machine 0 sees a and b in the dir listing,
    # but b does not really exist (open fails).
    xmkdir(0, "d$$");
    xcreate(0, "d$$/a", "11");
    xcreate(1, "d$$/b", "22");
    xls(0, "d$$", "a", "b");
    xmv(1, "d$$/b", "d$$/c");
    xls(0, "d$$", "a", "c");
    xrm(1, "d$$/a");
    xls(0, "d$$", "c");
    xrm(0, "d$$/c");
    xrmdir(0, "d$$");

    # nested mkdir, rmdir
    xmkdir(0, "de$$");
    xmkdir(0, "de$$/df$$");
    xcreate(0, "de$$/df$$/fh$$", "abcd");
    xcheckf(0, "de$$/df$$/fh$$", "abcd");
    xcheckf(1, "de$$/df$$/fh$$", "abcd");
    xrm(0, "de$$/df$$/fh$$");
    xrmdir(0, "de$$/df$$");
    xrmdir(0, "de$$");
    xcreate(0, "de$$/fg$$", "subone");
    xchecknot(0, "de$$/fg$$");
    xcreate(1, "de$$/fi$$", "subtwo");
    xchecknot(1, "de$$/fi$$");
}

# Remote file system operations.
# In general these x routines return undef on error.

# Make sure ls output contains what we expect.
sub xls {
    my($un, $d, @l) = @_;

    my $out;
    my $st = ssh($un, "/bin/ls $fs/$d", undef, \$out);
    my @a = split(/\n/, $out);
    @a = sort { $a cmp $b } @a;
    @l = sort { $a cmp $b } @l;

    my $i;
    my $ok = 1;
    for($i = 0; $i <= $#l; $i++){
        if($a[$i] ne $l[$i]){
            $ok = 0;
        }
    }
    if($ok != 1 || $#a != $#l){
        print STDERR "twotest.pl: FAILED: ls $d produced wrong output\n";
        print STDERR "  wanted : @l\n";
        print STDERR "  got    : @a\n";
        # exit(1);
    }
}

# Does a file have the right number of links?
sub xlinks {
    my($un, $file, $wanted) = @_;

    my $out;
    my $st = ssh($un, "/bin/cat $fs/$file > /dev/null ; /bin/ls -ld $fs/$file", undef, \$out);
    if($out =~ /[-drwx]+ +([0-9]+) /){
        if($1 != $wanted){
            fail("wanted $wanted links for $file, got $1");
        }
    } else {
        fail("could not find links for $file");
    }
}

sub xcreate {
    my($un, $file, $contents) = @_;

    if($sleeps){
        sleep(1);
    }

    my $out;
    my $st = ssh($un, "unset noclobber; /bin/cat > $fs/$file", $contents, \$out);
    if($st == 0){
        return 1;
    } else {
        return undef;
    }
}

sub xread {
    my($un, $file) = @_;

    my $out;
    my $st = ssh($un, "/bin/cat $fs/$file", undef, \$out);
    if($st == 0){
        return $out;
    } else {
        return undef;
    }
}

sub xtruncate {
    my($un, $file, $len) = @_;

    if($sleeps){
        sleep(1);
    }

    my $st = ssh($un, "/usr/bin/truncate -c -s $len $fs/$file", undef, undef);
}

sub xmv {
    my($un, $old, $new) = @_;

    if($sleeps){
        sleep(1);
    }

    my $st = ssh($un, "/bin/mv $fs/$old $fs/$new", undef, undef);
    return($st == 0 ? 1 : undef);
}

sub xln {
    my($un, $old, $new) = @_;

    if($sleeps){
        sleep(1);
    }

    my $st = ssh($un, "/bin/ln $fs/$old $fs/$new", undef, undef);
    return($st == 0 ? 1 : undef);
}

sub xsymlink {
    my($un, $old, $new) = @_;

    if($sleeps){
        sleep(1);
    }

    my $st = ssh($un, "/bin/ln -s $fs/$old $fs/$new", undef, undef);
    return($st == 0 ? 1 : undef);
}

sub xreadlink {
    my($un, $f, $wanted) = @_;

    my $out;
    my $st = ssh($un, "/bin/ls -ld $fs/$f", undef, \$out);
    if($st == 0 && $out =~ /-> ([^\n]+)/){
        if("$fs/$wanted" eq $1){
            # OK
            return;
        }
        fail("readlink $f, wanted $wanted, got $1");
    }
    fail("readlink $f failed");
}

sub xmkdir {
    my($un, $dir) = @_;

    if($sleeps){
        sleep(1);
    }

    my $st = ssh($un, "/bin/mkdir $fs/$dir", undef, undef);
    return($st == 0 ? 1 : undef);
}

sub xrmdir {
    my($un, $dir) = @_;

    if($sleeps){
        sleep(1);
    }

    my $st = ssh($un, "/bin/rmdir $fs/$dir", undef, undef);
    return($st == 0 ? 1 : undef);
}

sub xrm {
    my($un, $f) = @_;

    if($sleeps){
        sleep(1);
    }

    my $st = ssh($un, "/bin/rm $fs/$f", undef, undef);
    return($st == 0 ? 1 : undef);
}

sub xcheckf {
    my($un, $file, $wanted) = @_;
    
    if($sleeps){
        sleep(1);
    }

    my $x = xread($un, $file);
    if(!defined($wanted) && !defined($x)){
        # OK
    } elsif(!defined($wanted) || !defined($x) || $wanted ne $x){
        fail("content check failed, $file, $wanted, $x");
    }
}

# check that a file does not exist -- cannot be opened
sub xchecknot {
    my($un, $file) = @_;

    if($sleeps){
        sleep(1);
    }

    my $st = ssh($un, "/bin/cat $fs/$file", undef, undef);
    if($st == 0){
        fail("file exists but should not: $file");
    }
}

sub ssh {
    my($un, $cmd, $input, $output) = @_;

    my $if = "/tmp/iftt$$";
    my $of = "/tmp/oftt$$";

    open(IF, ">$if") or die "cannot write $if";
    if(defined($input)){
        print IF $input;
    }
    close(IF);
    unlink($of);

    my $h = $where->[$un][0];
    my $u = $where->[$un][1];
    my $c = "ssh -x -a -l $u $h '$cmd' < $if > $of";

    my $status = system($c);

    open(OF, $of) or die "cannot read $of";
    my $z = "";
    sysread(OF, $z, 32768);
    close(OF);

    unlink($if);
    unlink($of);

    if(defined($output)){
        $$output = $z;
    }
    return $status;
}

sub fail {
    my($msg) = @_;

    print STDERR "twotest.pl: FAILED: $msg\n";
#    exit(1);
}

main();
